#!/usr/bin/perl

# KoL CLI v0.98e
#
# A command-line interface for playing KoL

# CREDITS:
# Most code by and (c) 2004 Paul Lujan
# Based on original code by Marshall Perrin

# This is free software. You may copy and redistribute it under the
# same terms as Perl itself.

# Thanks to people who helped me out:
#   PumpkinEve, Ohayou, badfeet, greyselke, AmberDragon, Squiffle,
#   Allessindra, eric9, Catchy User Name, phenyx, nithan, tyger,
#   Stoop, Mustachio

# Latest changes:
# Version 0.98e:
# - Fixed fighting plural objects. Thanks to Mustachio.
# - Exact matches now work.
# - Doesn't do anything if the number specified is 0.
# - Added code for multiple custom outfits. Thanks to tyger for the basic
#   code.
# - New noodle/scrumptious reagent code
# - Arches and toast code updated. Note that the old code still works
#   but I figure it's much more polite this way.
# - When your bartender blows up, it shows you what parts you got.
# - The password hash is now obtained from the Demon Market.

# User-sent features I need to add:
# - badfeet's food/drink lists
# - nithan's inventory stuff
# - shoop's equip stuff

# Higher-priority:
# - more sophisticated macro arguments
# - use # to specify item number in send (or elsewhere for that matter)
# - Combine adventure results across multiple adventures in 1 session.
# - Fix battle parsing for more flavorful text (e.g. "It entangles..." in bat hole)
# - Fix battle parsing when, e.g., potato entangles enemy.
# - Finish mix verb (and add cook, etc.)
# - add in autodetect of boot back to map screen during fights to prevent loops
# - finish skills for PMs
# - add in "use all" option
# - casino
# - compact battle results mode (suppress familiar output too)
# - properly implement, test, and foolproof skill use in battle
# - (re-)implement healevery feature
# - ability to change the server you're on
# - Sewer Luck

# Medium-priority:
# - inventory argument
# - autoequip cummerbund for mafia aria, talisman for Palindome, etc.
# - allow user to set which outfit they want equipped in autoequip
# - have adventure summary include stats as well as items
# - track MP on client side too
# - fix adventure titles in limerick dungeon
# - add in verbose functionality (and DECREASED verbosity)
# - kick into verbose mode if parsing fails in do_action
# - add cocktail-crafting/food-cooking abilities?
# - advanced buying (e.g. equip hippy outfit to buy herbs, etc.)?
# - detect combat at beginning of session
# - have ^Cs interrupt current operation rather than stopping entirely
# - revamp item system to allow colossal closet/clan stash/mall store operations

# Lowest-priority (some unimportant, some likely to be time-consuming):
# - Better message system (e.g. read messages too)?
# - Daily dungeon
# - Fullness meter
# - Readline interface
# - Ohayou's parser suggestion
# - tweak random adventure location
# - Some kind of external log?

# BUGS (NOT MY FAULT):
#  - When vacationing at the shore, gains of points aren't always displayed.
#  - In compact mode, we get the familiar weight rather than its name.

use strict;

my $version="0.98e";

# Usernames, passwords, password hashes, and settings
# read in from the config file
my @users;
my %passwords;
my %settings;        # actually a double hash -- $settings{$user}{$param} == value

# Variables for interacting with server
my @response;        # Last response from server
my $response="";     # same in string form
my $username;
my $password;
my $pwdhash;
my $cookiefilename;

# Preferences
my $nturns=1;        # Default number of turns to perform an action
my $healevery=0;     # How often to heal oneself in combat
my $n=0;             # How much worthless stuff we got in the sewer
my $cloverprotect=1; # How zealously to protect our clovers
my $debug=0;         # Debug mode
my $verbose=0;       # Verbose mode
my $noconnect=0;     # Don't actually connect to the web site; just test the program internals
my $macrofail=1;     # How aggressively to continue if our macro hits a glitch
my $alwaysequip=0;   # Whether or not to always equip our outfit.
my $summary=1;       # Whether or not to display the adventure summary.
my $maxdrunk=20;     # The maximum drunkenness we can tolerate.
my $fleewhen=0;      # Flee when your HP are less than or equal to this value.
my $checkbuffbots=1; # Whether or not to check if the buffbots are online.

# Status of our session
my $relogins=0;      # Number of times we've relogged in. Protects against infinite loops & such.
my $currentoutfit=-99;  # The outfit we're currently wearing.
my $defaultoutfit="custom";
my ($curhp, $maxhp); # Current/maximum HP: Not actually used but they're set for now in case I
my ($curmp, $maxmp); # Current/maximum MP: want to implement them later.
my $advleft;         # Adventures remaining.
my $drunk;           # Current drunkenness
my $hassphere;       # If we have an elemental saucesphere. Only used for the Peak for now.
my $famname;         # Familiar name.
my $famweight;       # Familiar weight. If in compact mode, only this one will be filled.
my $famtype;         # Familiar type.
my $buffbots;        # If the buffbots are online
my %customoutfits;   # Hash containing our list of custom outfits.
my $gotoutfits=0;    # If we've already pulled up the list of custom outfits or not.
# Note that we can't rely on %customoutfits being empty to do this since it's possible
# the user has no custom outfits.

# Program internal status
my $object;           # Current object in parsed line. This is global so do_action can grab it
                      # in the mall fallback code.
my $allowedobjects;   # Reference to allowed objects hash. This is global so we can just have
                      # it set once by get_allowedobjects.
my $batchmode=0;      # if we were run in batch mode
my $macroexecuting=0; # if a macro is currently executing
my $macroerror=0;     # if there's an error in the current macro
my @currentmacro;     # what the current macro is
my $prevverb;         # previous command
my @prevwords;        # rest of previous line
my %macros;           # List of macros read in from config file
my $onquest=0;        # Whether or not we're on a quest
my $questobj;         # Object to obtain on quest
my $questnum;         # Number of said objects to obtain
my $questcur;         # Number of said objects obtained so far
my $questmax=0;       # Maximum number of adventures to spend questing
my @questobjs;        # List of allowed objects for questing. Not set here
                      # but you can create your own in extensions.kol.
my %advitems;         # Items we got on our adventure.
my %advgains;         # Stats and Meat we got on our adventure.

# curl parameters
my $site = "http://www.kingdomofloathing.com";
# spoof user-agent for now. Later, maybe not.
my $ua = "-A \"Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.6) Gecko/20040413 Debian/1.6-5\"";
my $cookies = "-b .kolcookies-USERNAME -c .kolcookies-USERNAME"; # store persistent cookies here
my $options;

# Check to make sure curl actually exists.
my $curlversion = `curl --version` or die "Couldn't find curl. Please make sure that curl is installed and in your \$PATH.\n";

# Data.
# List of possible actions.

# To completely add in a new verb:
# (1) Add it here
# (2) Add in a help line in %verbhelp
# (3) Create its list of objects
# (4) Point the parser loop in &get_allowedobjects to the list of objects
# (5) If it's not a standard action (i.e. goes to do_action), put the appropriate
#     call at the end of the main parser loop
# (6) If the help is not standard, put in something appropriate in &do_help

my %verbs = ("adventure" => "adventure.php", "buff" => "skills.php", "buy" => "store.php",
	     "campground" => "campground.php", "clan" => "clan_stash.php", "donate" => "shrines.php",
	     "drink" => "inv_booze.php", "eat" => "inv_eat.php", "equip" => "inv_equip.php", 
	     "execute" => "dummy", "exit" => "logout.php", "familiar" => "familiar.php",
	     "getbuff" => "sendmessage.php", "gym" => "clan_gym.php", "help" => "dummy",
	     "hermit" => "hermit.php", "junk" => "sellstuff.php", "logout" => "logout.php",
	     "mallbuy" => "mallstore.php", "mix" => "cocktail.php", "moons" => "compactmenu.php",
	     "outfit" => "inv_equip.php", "repeat" => "dummy", "set" => "dummy", "send" => "sendmessage.php", 
	     "shore" => "shore.php", "skill" => "skills.php", "status" => "charpane.php",
	     "switch" => "dummy", "togglemode" => "account.php", "trinket" => "adventure.php",
	     "quit" => "logout.php", "use" => "inv_use.php", "halloween" => "trickortreat.php");

my %verbhelp = ("adventure" => "adventure <loc> [<n>]: Adventure in location <loc> for <n> turns.",
		"buff" => "buff <buff> <target> [<n>]: Give <target> <n> turns of buff <buff>.",
		"buy" => "buy <item> [<n>]: Buy <n> of <item> from Market Square.",
		"campground" => "campground <action> [<n>]: Perform <action> in your campground <n> times.",
		"clan" => "clan <action> [<n>]: Perform <action> in your clan hall <n> times.",
		"donate" => "donate <shrine> [<n>]: Donate <n> meat to the shrine of <shrine>.",
		"drink" => "drink <drink> [<n>]: Drink <n> of <drink>.",
		"eat" => "eat <food> [<n>]: Eat <n> of <food>.",
		"equip" => "equip <item>: Equip <item> on your character.",
		"execute" => "execute <macro> [<arg>]: Execute <macro> with argument <arg>.",
		"exit" => "exit: End your session.",
		"familiar" => "familiar <familiar>: Bring along familiar <familiar> (use \"list\" to list all).",
		"getbuff" => "getbuff <buff> [<n>]: Use the buffbots to obtain buff <buff> <n> times.",
		"gym" => "gym <activity> [<n>]: Work out in the clan gym doing <activity> <n> times.",
		"halloween" => "halloween trickortreat [<n>]: Go trick-or-treating <n> times.",
		"help" => "help [<command>]: Display help (for <command> if specified).",
		"hermit" => "hermit <item> [<n>]: Obtain item <item> from the hermit <n> times.",
		"junk" => "junk <area>: Sell all junk items you obtained from area <area>.",
		"logout" => "logout: End your session.",
		"mallbuy" => "mallbuy <item> [<n>]: Buy <n> of <item> from a few select Mall stores.",
		"mix" => "mix <drink> [<n>]: Mix <n> of <drink> from ingredients.",
		"moons" => "moons: Display moon phase and any special effects due to them.", 
		"outfit" => "outfit <outfit>: Equip <outfit>.",
		"quit" => "quit: End your session.",
		"repeat" => "repeat: Repeats the previous command.",
		"send" => "send <item> <target> [<n>]: Send <n> of <item> to <target>.",
		"set" => "set <parameter> <value>: Set <parameter> to <value>.\n  If no parameter is specified, displays the current value of all parameters.",
		"shore" => "shore <destination> [<n>]: Take a vacation to <destination> <n> times.",
		"skill" => "skill <skill> [<n>]: Use skill <skill> <n> times on yourself.",
		"status" => "status: Display your character's status.",
		"switch" => "switch <user>: Log out your current character and log in with <user>.",
		"togglemode" => "togglemode: Switch between compact and full mode.",
		"trinket" => "trinket [<n>]: Obtain <n> trinkets from The Sewer.",
		"use" => "use <item> [<n>]: Use <item> <n> times.");

# List of supported objects for each verb.

my %areas = ("sleazybackalley" => 9, "hauntedpantry" => 10, "sewer" => 12,            # town
	     "haikudungeon" => 16, "limerickdungeon" => 19, "doom" => 39,             # dungeon full of dungeons

	     "koutskirts" => 11, "kkitchens" => 40, "ktreasury" => 41, "kharem" => 42,  # cobb's knob
	     "klaboratory" => 50, "km1" => 51, "km2" => 52, "km3" => 53,                # cobb's knob lab

	     "degrassi" => 18, "funhouse" => 20, "fernswarthy" => 22,                 # plains
	     "cemetary" => 58, # is this different before/after cyrpt? if so, we'll need to get the earlier
	                       # adventure number too
	                                                                             
	     "bhentry" => 30, "bhguano" => 31, "bhrat" => 32, "bhbean" => 33,         # bat hole

	     "spookyforest" => 15, "hiddentemple" => 17, "typicaltavern" => 25,       # woods
	     "whiteysgrove" => 28, "eightbit" => 73, "dffgate" => 79,                 # more woods

	     "barrel" => 35, "valley" => 80,                                          # mountains
	     "goatlet" => 60, "mine" => 61, "ninjasnowmen" => 62, "extreme" => 63,    # mt. mclargehuge
	     "icypeak" => 64, "dmine" => 36,                                          # more

	     # Technically speaking, we don't have to feed the correct number
	     # since it'll determine whether we're in disguise or not. However,
	     # since otherwise we have no way to tell whether the user wants to
	     # be disguised or not, we might as well be polite.

	     "hippycamp" => 26, "frathouse" => 27, "piratecove" => 66,                # island
	     "dhippycamp" => 65, "dfrathouse" => 29, "dpiratecove" => 67,             # island in disguise

	     "picklefactory" => 37, "palindome" => 78,                                # special plains
	     "airship" => 81, "giantcastle" => 82, "holeinthesky" => 83,              # beanstalk
	     "stupor" => 23, "random" => -1, "arandom" => -2);

# Mapping of area numbers to area names. This is mostly for aesthetic
# purposes.

my %areanames = (9 => "Sleazy Back Alley", 10 => "Haunted Pantry", 12 => "Sewer",
		 16 => "Haiku Dungeon", 19 => "Limerick Dungeon", 39 => "Dungeons of Doom",

		 11 => "Cobb's Knob Outskirts", 40 => "Cobb's Knob Kitchens", 41 => "Cobb's Knob Treasury",
		 42 => "Cobb's Knob Harem", 50 => "Knob Laboratory", 51 => "Menagerie Level 1",
		 52 => "Menagerie Level 2", 53 => "Menagerie Level 3",

		 18 => "Degrassi's Knoll", 20 => "'Fun' House", 22 => "Fernswarthy's Tower",
		 58 => "Misspelled Cemetary", 
		 30 => "Bat Hole Entryway", 31 => "Guano Junction", 32 => "Batrat and Ratbat Burrow",
		 33 => "Beanbat Chamber",
		 
		 15 => "Spooky Forest", 17 => "Hidden Temple", 25 => "Typical Tavern",
		 28 => "Whitey's Grove", 73 => "8-Bit Realm", 79 => "Deep Fat Friar's Gate",

		 35 => "Barrel Full of Barrels", 80 => "Valley of Rof L'm Fao",
		 60 => "Goatlet", 61 => "Itznotyerzitz Mine", 62 => "Lair of the Ninja Snowmen",
		 63 => "eXtreme Slope", 64 => "Icy Peak", 36 => "Itznotyerzitz Mine (in disguise)",
		 
		 26 => "Hippy Camp", 27 => "Orcish Frat House", 66 => "Obligatory Pirate's Cove",
		 65 => "Hippy Camp (in disguise)", 29 => "Orcish Frat House (in disguise)",
		 67 => "Obligatory Pirate's Cove (in disguise)",

		 37 => "Spectral Pickle Factory", 78 => "Palindome",

		 81 => "Penultimate Fantasy Airship", 82 => "Giant Castle", 83 => "Hole in the Sky",

		 23 => "Drunken Stupor",
		 -1 => "Random Area (restricted)", -2 => "Random Area (all)");

# Big ol' hashes with the various form parameters for items.
# This would probably be more efficiently done, in many cases, by having one single
# template and a hash containing only the item number. Maybe this will happen
# someday. For now, I'm staying with the easy way.

my %buyobjs = ("gum" => "whichstore=m&buying=Yep.&phash=PASSWORDHASH&whichitem=23&howmany=QTY",
	       "fermenting" => "whichstore=m&buying=Yep.&phash=PASSWORDHASH&whichitem=247&howmany=QTY",
	       "hairspray" => "whichstore=m&buying=Yep.&phash=PASSWORDHASH&whichitem=744&howmany=QTY",
	       "dgunguent" => "whichstore=d&buying=Yep.&phash=PASSWORDHASH&whichitem=231&howmany=QTY",
	       "dgointment" => "whichstore=d&buying=Yep.&phash=PASSWORDHASH&whichitem=232&howmany=QTY",
	       "dgbalm" => "whichstore=d&buying=Yep.&phash=PASSWORDHASH&whichitem=233&howmany=QTY",
	       "dgelixir" => "whichstore=d&buying=Yep.&phash=PASSWORDHASH&whichitem=234&howmany=QTY");

my %campgroundobjs = ("toast" => "action=toast", "evilfood" => "action=arches", "rest" => "action=rest",
		      "relax" => "action=relax");

my %clanobjs = ("donate" => "action=contribute&howmuch=QTY",
		"cmuscle" => "action=buyround&size=1&whichgift=1",
		"nmuscle" => "action=buyround&size=2&whichgift=1",
		"emuscle" => "action=buyround&size=3&whichgift=1",
		"cmysticality" => "action=buyround&size=1&whichgift=2",
		"nmysticality" => "action=buyround&size=2&whichgift=2",
		"emysticality" => "action=buyround&size=3&whichgift=2",
		"cmoxie" => "action=buyround&size=1&whichgift=3",
		"nmoxie" => "action=buyround&size=2&whichgift=3",
		"emoxie" => "action=buyround&size=3&whichgift=3",
		"ctmuscle" => "action=buyround&size=1&whichgift=4",
		"ntmuscle" => "action=buyround&size=2&whichgift=4",
		"etmuscle" => "action=buyround&size=3&whichgift=4",
		"ctmysticality" => "action=buyround&size=1&whichgift=5",
		"ntmysticality" => "action=buyround&size=2&whichgift=5",
		"etmysticality" => "action=buyround&size=3&whichgift=5",
		"ctmoxie" => "action=buyround&size=1&whichgift=6",
		"ntmoxie" => "action=buyround&size=2&whichgift=6",
		"etmoxie" => "action=buyround&size=3&whichgift=6");

my %donateobjs = ("muscle" => "pwd=PASSWORDHASH&action=boris&howmuch=QTY",
		  "boris" => "pwd=PASSWORDHASH&action=boris&howmuch=QTY",
		  "mysticality" => "pwd=PASSWORDHASH&action=jarlsberg&howmuch=QTY",
		  "jarlsberg" => "pwd=PASSWORDHASH&action=jarlsberg&howmuch=QTY",
		  "moxie" => "pwd=PASSWORDHASH&action=sneakypete&howmuch=QTY",
		  "sneakypete" => "pwd=PASSWORDHASH&action=sneakypete&howmuch=QTY");

my %drinkobjs = ("bloodybeer" => "which=1&whichitem=252", "rollinthehay" => "which=1&whichitem=679",
		 "saltydog" => "which=1&whichitem=248", "esbm" => "which=1&whichitem=257",
		 "finewine" => "which=1&whichitem=255", "schlitz" => "which=1&whichitem=41",
		 "willer" => "which=1&whichitem=81", "impale" => "which=1&whichitem=470",
		 "fuzzbump" => "which=1&whichitem=799", "grapefruit" => "which=1&whichitem=254",
		 "orange" => "which=1&whichitem=253", "tomato" => "which=1&whichitem=256",
		 "dirtymartini" => "which=1&whichitem=948");
		   
my %eatobjs = ("delicious" => "which=1&whichitem=573", "abominable" => "which=1&whichitem=639",
	       "isbb" => "which=1&whichitem=316", "isebb" => "which=1&whichitem=319",
	       "ramen" => "which=1&whichitem=578", "tofucm" => "which=1&whichitem=816",
	       "batwingcm" => "which=1&whichitem=814", "knobcm" => "which=1&whichitem=813",
	       "proncm" => "which=1&whichitem=817", "ratcm" => "which=1&whichitem=815",
	       "bklp" => "which=1&whichitem=513", "jklp" => "which=1&whichitem=514",
	       "spklp" => "which=1&whichitem=515", "pickle" => "which=1&whichitem=764",
	       "tomato" => "which=1&whichitem=246", "spizza" => "which=1&whichitem=324");

my %equipobjs = ("talisman" => "which=2&action=equip&whichitem=486");

my %getbuffobjs = ("ode" => "action=send&pwd=PASSWORDHASH&towho=95114&sendmeat=2001",
		   "moxious" => "action=send&pwd=PASSWORDHASH&towho=95114&sendmeat=2004",
		   "polka" => "action=send&pwd=PASSWORDHASH&towho=95114&sendmeat=4007",
		   "mojomuscular" => "action=send&pwd=PASSWORDHASH&towho=95114&sendmeat=6008",
		   "arrowsmith" => "action=send&pwd=PASSWORDHASH&towho=95114&sendmeat=7010",
		   "phatloot" => "action=send&pwd=PASSWORDHASH&towho=95114&sendmeat=12013",
		   "superiority" => "action=send&pwd=PASSWORDHASH&towho=95114&sendmeat=20017",
		   "empathy" => "action=send&pwd=PASSWORDHASH&towho=98842&sendmeat=10012",
		   "tenacity" => "action=send&pwd=PASSWORDHASH&towho=98842&sendmeat=16016",
		   "astral" => "action=send&pwd=PASSWORDHASH&towho=98842&sendmeat=14014",
		   "reptilian" => "action=send&pwd=PASSWORDHASH&towho=98842&sendmeat=14015",
		   "ghostly" => "action=send&pwd=PASSWORDHASH&towho=98842&sendmeat=8011");

my %gymobjs = ("muscle" => "action=hoboflex&numturns=QTY", "hoboflex" => "action=hoboflex&numturns=QTY",
	       "mysticality" => "action=bigbook&numturns=QTY", "bigbook" => "action=bigbook&numturns=QTY",
	       "moxie" => "action=tanningbed&numturns=QTY", "tanulots" => "action=tanningbed&numturns=QTY");

my %halloweenobjs = ("trickortreat" => "action=Yep.");

my %mallobjs = ("elixir" => "pwd=PASSWORDHASH&whichstore=125017&buying=Yep.&whichitem=277000000700&quantity=QTY",
		"jug" => "pwd=PASSWORDHASH&whichstore=125017&buying=Yep.&whichitem=278000000700&quantity=QTY",
		"ramen" => "pwd=PASSWORDHASH&whichstore=91072&buying=Yep.&whichitem=578000008999&quantity=QTY",
		"rum" => "pwd=PASSWORDHASH&whichstore=43579&buying=Yep.&whichitem=787000002500&quantity=QTY",
		"gin" => "pwd=PASSWORDHASH&whichstore=43579&buying=Yep.&whichitem=237000002000&quantity=QTY",
		"olive" => "pwd=PASSWORDHASH&whichstore=43579&buying=Yep.&whichitem=245000000150&quantity=QTY");

my %mixobjs = ("martini" => "action=combine&pwd=PASSWORDHASH&item1=237&item2=245&quantity=QTY",
	       "skolive" => "action=combine&pwd=PASSWORDHASH&item1=938&item2=245&quantity=QTY",
	       "dirtymartini" => "action=combine&pwd=PASSWORDHASH&item1=251&item2=945&quantity=QTY");

# Note: If you create your own sendobjs and you want them to work with the ability to put the message
# text in the command line, make sure that they end with the "&message=" part.

my %sendobjs = ("meat" => "action=send&pwd=PASSWORDHASH&towho=NAME&sendmeat=QTY&message=",
		"beer" => "action=send&pwd=PASSWORDHASH&towho=NAME&whichitem1=41&howmany1=QTY&message=");

my %shoreobjs = ("muscle" => "pwd=PASSWORDHASH&whichtrip=1", "ranch" => "pwd=PASSWORDHASH&whichtrip=1",
		 "mysticality" => "pwd=PASSWORDHASH&whichtrip=2", "island" => "pwd=PASSWORDHASH&whichtrip=2",
		 "moxie" => "pwd=PASSWORDHASH&whichtrip=3", "ski" => "pwd=PASSWORDHASH&whichtrip=3");

# Buffs are still duplicated between %buffobjs and %skillobjs, but at least it's a lot
# easier to read this way.

my %buffobjs = ("ghostly" => "pwd=PASSWORDHASH&action=Skillz&whichskill=2007&specificplayer=NAME&bufftimes=QTY",
		"fortitude" => "pwd=PASSWORDHASH&action=Skillz&whichskill=2008&specificplayer=NAME&bufftimes=QTY",
		"empathy" => "pwd=PASSWORDHASH&action=Skillz&whichskill=2009&specificplayer=NAME&bufftimes=QTY",
		"tenacity" => "pwd=PASSWORDHASH&action=Skillz&whichskill=2010&specificplayer=NAME&bufftimes=QTY",
		"astral" => "pwd=PASSWORDHASH&action=Skillz&whichskill=2012&specificplayer=NAME&bufftimes=QTY",
		"elemental" => "pwd=PASSWORDHASH&action=Skillz&whichskill=4007&specificplayer=NAME&bufftimes=QTY",
		"jalapeno" => "pwd=PASSWORDHASH&action=Skillz&whichskill=4008&specificplayer=NAME&bufftimes=QTY",
		"jabanero" => "pwd=PASSWORDHASH&action=Skillz&whichskill=4011&specificplayer=NAME&bufftimes=QTY",
		"healthiness" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6003&specificplayer=NAME&bufftimes=QTY",
		"moxious" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6004&specificplayer=NAME&bufftimes=QTY",
		"celerity" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6005&specificplayer=NAME&bufftimes=QTY",
		"polka" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6006&specificplayer=NAME&bufftimes=QTY",
		"mojomuscular" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6007&specificplayer=NAME&bufftimes=QTY",
		"arrowsmith" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6008&specificplayer=NAME&bufftimes=QTY",
		"absorption" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6009&specificplayer=NAME&bufftimes=QTY",
		"phatloot" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6010&specificplayer=NAME&bufftimes=QTY",
		"pointiness" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6011&specificplayer=NAME&bufftimes=QTY",
		"destruction" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6012&specificplayer=NAME&bufftimes=QTY",
		"superiority" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6013&specificplayer=NAME&bufftimes=QTY",
		"ode" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6014&specificplayer=NAME&bufftimes=QTY");

my %skillobjs = (%buffobjs,
		 "sealclubbing" => "pwd=PASSWORDHASH&action=Skillz&whichskill=1000&quantity=QTY",
		 "ottertongue" => "pwd=PASSWORDHASH&action=Skillz&whichskill=1007&quantity=QTY",
		 "walrustongue" => "pwd=PASSWORDHASH&action=Skillz&whichskill=1010&quantity=QTY",
		 "patience" => "pwd=PASSWORDHASH&action=Skillz&whichskill=2000&quantity=QTY",
		 "manicotti" => "pwd=PASSWORDHASH&action=Skillz&whichskill=3000&quantity=QTY",
		 "noodles" => "pwd=PASSWORDHASH&action=Skillz&whichskill=3006&quantity=QTY",
		 "lasagna" => "pwd=PASSWORDHASH&action=Skillz&whichskill=3009&quantity=QTY",
		 "leash" => "pwd=PASSWORDHASH&action=Skillz&whichskill=3010&quantity=QTY",
		 "cannelloni" => "pwd=PASSWORDHASH&action=Skillz&whichskill=3012&quantity=QTY",
		 "saucecontemp" => "pwd=PASSWORDHASH&action=Skillz&whichskill=4000&quantity=QTY",
		 "scrumptious" => "pwd=PASSWORDHASH&action=Skillz&whichskill=4006&quantity=QTY",
		 "daerobics" => "pwd=PASSWORDHASH&action=Skillz&whichskill=5000&quantity=QTY",
		 "dnap" => "pwd=PASSWORDHASH&action=Skillz&whichskill=5007&quantity=QTY",
		 "dpowernap" => "pwd=PASSWORDHASH&action=Skillz&whichskill=5011&quantity=QTY",
		 "mariachi" => "pwd=PASSWORDHASH&action=Skillz&whichskill=6000&quantity=QTY");

# MP costs of all of these things. Thanks to Ohayou for most of this data.

my %mpcosts = ("sealclubbing" => 1, "ottertongue" => 10, "walrustongue" => 10, "patience" => 1,
	       "ghostly" => 6, "fortitude" => 10, "empathy" => 15, "tenacity" => 12, "astral" => 10,
	       "manicotti" => 1, "lasagna" => 5, "leash" => 12, "cannelloni" => 16, "saucecontemp" => 1,
	       "elemental" => 3, "jalapeno" => 5, "jabanero" => 10, "daerobics" => 1, "dnap" => 7,
	       "dpowernap" => 10, "mariachi" => 1, "healthiness" => 3, "moxious" => 4, "celerity" => 6,
	       "polka" => 8, "mojomuscular" => 11, "arrowsmith" => 15, "absorption" => 19,
	       "phatloot" => 24, "pointiness" => 29, "destruction" => 35, "superiority" => 42,
	       "ode" => 50);

# %useobjs is divided into two categories: items which can be Use Multipled and those which can't.
# First the can'ts.

my %useobjs = ("pastasummon" => "which=3&whichitem=780", 
	       "dough" => "which=3&whichitem=159", "flatdough" => "which=3&whichitem=301",
	       "yetisign" => "which=3&whichitem=775", "raffleticket" => "which=3&whichitem=500",
	       "mafiaaria" => "which=3&whichitem=781", "redppotion" => "which=3&whichitem=464",
	       "wereseal" => "which=3&whichitem=687", "healing" => "which=3&whichitem=595",

# Can be multiply used.

	       "elixir" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=277",
	       "jug" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=278",
	       "lotion" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=279",
	       "clover" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=24",
	       "disassembled" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=196",
	       "drinktickets" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=400",
	       "sixpack" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=138",
	       "largebox" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=832",
	       "smallbox" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=831",
	       "pfantasychest" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=604",
	       "grabbag" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=401",
	       "warmsubject" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=621",
	       "blueppotion" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=465",
	       "greenppotion" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=466",
	       "mmj" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=518",
	       "tinyhouse" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=592",
	       "phonicsdown" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=593",
	       "asjuice" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=794",
	       "arseafire" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=418",
	       "clemonade" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=424",
	       "dillpotion" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=767",
	       "drabpotion" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=422",
	       "equilibrium" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=425",
	       "gjuiceofpower" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=420",
	       "orangish" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=419",
	       "chalkyequality" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=793",
	       "gr8tness" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=426",
	       "tomatosmoothie" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=423",
	       "wrathserum" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=421",
	       "papower" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=926",
	       "micks" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=618",
	       "hairspray" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=744",
	       "spikyhairgel" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=587",
	       "dgunguent" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=231",
	       "dgointment" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=232",
	       "dgbalm" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=233",
	       "dgelixir" => "action=useitem&pwd=PASSWORDHASH&quantity=QTY&whichitem=234");

# Hermit items operate slightly differently -- we don't need the full list of parameters since the
# hermit has his own routine.

my %hermititems = ("clover" => 24, "figurine" => 46, "roll" => 47, "banjo" => 52, "pepper" => 55,
		   "cookie" => 61, "twig" => 66, "ketchup" => 106, "catsup" => 107, "rims" => 135,
		   "planks" => 140, "volleyball" => 527);

# Similarly for outfits and familiars.

my %outfitobjs = ("bugbear" => 1, "hippy" => 2, "fratboy" => 3, "haremgirl" => 4, "eliteguard" => 5,
		  "ninjasuit" => 6, "coldweather" => 7, "mining" => 8, "swashbuckling" => 9, "crimbo" => 11,
		  "stargarb" => 12, "furry" => 14, "eightbit" => 15, "yendorian" => 16);

my %familiarobjs = ("none" => -1, "list" => 0, "mosquito" => 1, "leprechaun" => 2, "potato" => 3, "goat" => 4,
		    "lime" => 5, "dice" => 6, "pirate" => 7, "barrrnacle" => 8, "monkey" => 9,
		    "bat" => 10, "grue" => 11, "volleyball" => 12, "ghuol" => 14, "fairy" => 15,
		    "cocoabo" => 16, "starfish" => 17, "pickle" => 19, "bee" => 20, "coffee" => 22, 
		    "cheshire" => 23, "jill" => 24, "turkey" => 25, "elfling" => 26, "dreidl" => 27,
		    "yeti" => 28);

# Backwards hash to translate familiar numbers back to familiar names.

my %familiarnames = (1 => "Mosquito", 2 => "Leprechaun", 3 => "Levitating Potato", 4 => "Angry Goat", 
		     5 => "Sabre-Toothed Lime", 6 => "Fuzzy Dice", 7 => "Spooky Pirate Skeleton",
		     8 => "Barrrnacle", 9 => "Howling Balloon Monkey", 10 => "Stab Bat",
		     11 => "Grue", 12 => "Blood-Faced Volleyball", 14 => "Ghuol Whelp",
		     15 => "Baby Gravy Fairy", 16 => "Cocoabo", 17 => "Star Starfish",
		     19 => "Ghost Pickle on a Stick", 20 => "Killer Bee", 22 => "Coffee Pixie",
		     23 => "Cheshire Bat", 24 => "Jill-O-Lantern", 25 => "Hand Turkey",
		     26 => "Crimbo Elf", 27 => "Hanukkimbo Dreidl", 28 => "Baby Yeti");

# Junk objects take a list of objects to be junked. These are arranged alphabetically
# unlike the areas themselves which are arrange by larger area. Clever, huh?

my %junkobjs = ("airship" => [ # Amulet of Extreme Plot Significance, cocoa eggshell fragment, glowing red eye, Metallic A
			       594, 589, 629, 628,
			       # Mohawk wig, photoprotoneutron torpedo, ridiculously huge sword, super-spikey hair gel
			       597, 630, 586, 587,
			       # titanium assault umbrella
			       596],
		"barrel" => [ # concentrated Magicalness pill, DG's Homeopathic Elixir, DG's Restorative Balm
			      272, 234, 233,
			      # Enchanted Barbell, giant Moxie weed, grapefruit, grapes, Mad Train wine, olive
			      271, 273, 243, 244, 564, 245],
		"bhbean" => [186], # enchanted bean -- though why you would go here if you didn't want those, I don't know
		"bhentry" => [188, 182], # bat guano, batgut
		"bhguano" => [ # bat guano, batgut, briefcase, broken skull, loose teeth, sonar-in-a-biscuit, wad of dough
			       188, 182, 184, 741, 187, 563, 159],
		"bhrat" => [ # bat guano, batgut, ratgut
			     188, 182, 562],
		"cemetary" => [ # dead guy's watch, ghuol ears, ghuol egg, grave robbing shovel,
				230, 166, 161, 671, 
				# lihc eye, lihc face, loose teeth, skeleton bone, spiked femur
				170, 670, 187, 163, 668],
		"degrassi" => [ # Bugbear Beanie, cog, gnoll lips, gnoll teeth, Gnollish flyswatter, Gnollish pie tin,
				169, 120, 172, 220, 123, 158, 117,
				# Gnollish plunger, ice-cold beer (Willer), skewer, sprocket, tires, wad of dough
				117, 81, 165, 119, 136, 159],
		"dffgate" => [ # cast, diamond-studded cane, Evil Golden Arch, flaming crutch, hot katana blade
			       474, 472, 478, 473, 350,
			       # hot wing, Imp Ale, infernal insoles, leather mask, ninja hot pants, Ruby W, wussiness potion
			       471, 477, 475, 352, 468, 469],
		"doom" => [# cornuthaum, mind flayer corpse, ring of aggravate monster, royal jelly, vorpal blade
			   834, 836, 835, 830, 833],
		"eightbit" => [ # black Pixel, blue Pixel, green Pixel, red Pixel
				460, 463, 462, 461], 
		"evilfood" => [ # Brimstone Chicken, Beelzeburger, Dr. Lucifer, Lord of the Flies-sized fries
				570, 568, 571, 569],
		"extreme" => [ # eXtreme mittens, eXtreme Scarf, sk8board, snowboarder pants, t8r tots
			       399, 355, 410, 356, 359],
		"fernswarthy" => [ # taco shell, uncooked chorizo, wad of dough
				  173, 175, 159],
		"frathouse" => [ # homoerotic frat-paddle, ice-cold beer, ice-cold fotie, Orc baseball cap, Orc cargo shorts
				241, 41, 180, 239, 240],
		"funhouse" => [ # acid-squirting flower, big red clown nose, bloody clown pants, Bugbear Beanie
				430, 449, 432, 169,
				# Bugbear Bungguard, clown shoes, disease, foolscap fool's cap
				79, 431, 452, 448],
		"giantcastle" => [ # Angry Farmer candy, awful poetry journal, disturbing fanfic, furry fur
				   617, 622, 632, 616,
				   # giant needle, heavy D, original G, plot hole, probability potion
				   619, 611, 612, 613, 614,
				   # procrastination potion, wolf mask
				   610, 633], 
		"goatlet" => [329, 334], # goat beard, sabre teeth
		"haikudungeon" => [ # Bugbear Bungguard, crusty L-shaped piece of metal, ice-cold beer (Willer), 
				    79, 83, 81,
				    # locked Orcish meat locker, meat from yesterday, rusty L-shaped piece of metal
				    84, 87, 82],
		"hauntedpantry" => [19, 25, 246], # asparagus knife, meat paste, tomato
		"hippycamp" => [ # decorative fountain, Feng Shui, filthy corduroys, filthy knitted dread sack
				 211, 210, 213, 214,
				 # hemp bracelet, herbs, necklace chain, patchouli incense stick, phat turquoise bead,
				 218, 203, 219, 208, 222, 206,
				 # piercing post, tambourine bells, windchimes
				 206, 739, 212],
		"holeinthesky" => [655, 654, 656], # line, star, star chart
		"icypeak" => [ # iron pasta spoon, kneecapping stick, Mob Penguin cellular phone, 
			       777, 776, 779,
			       # penguin skin, support cummerbund, yeti fur
			       393, 778, 388],
		"island" => [138, 139], # ice-cold six-pack, valuable trinket
		"kharem" => [# KG Elite Helm, KG Elite Pants, KG Elite Polearm, KG Perfume
			     308, 309, 310, 307], 
		"kkitchens" => [246, 159], # tomato, wad of dough
		"klaboratory" => [ # KG Seltzer, KG Steroids, KG Superseltzer
				   344, 340, 345],
		"km1" => [244, 242], # grapes, orange
		"km2" => [273, 14], # giant Moxie weed, Moxie weed
		"koutskirts" => [ # chef's hat, Dr. Hobo's map, KG pants, KG scimitar, KG tongs, Strongness elixir
				  293, 601, 38, 35, 36, 15,
				  # viking helmet
				  37],
		"ktreasury" => [ # KG Elite Helm, KG Elite Pants, KG Elite Polearm, KG Visor, meat stack
				 308, 309, 310, 312, 88],
		"limerickdungeon" => [296, 768, 771], # dungeoneer's dungarees, spork, tassel
		"mine" => [ # 7-Foot Dwarven Mattock, asbestos ore, linoleum ore, miner's helmet, miner's pants
			    362, 364, 363, 360, 361],
		"ninjasnowmen" => [ # cold ninja mask, frigid ninja stars, frozen nunchaku, icy katana hilt
				    348, 353, 354, 349],
		"palindome" => [ # ketchup hound, leather chaps, mesh cap, drab sonata, guitar pick, olive
				 493, 492, 490, 489, 488, 245,
				 # superolive, bat Wing, batblade, batgut, enormous belt buckle
				 455, 183, 494, 182, 491,
				 # cat appendix, catgut, taco shell, gnatwing 
				 496, 495, 173, 497],
		"piratecove" => [ # acoustic guitarrr, arrrgyle socks, barrnacle, charrrm, crowbarrr, eyepatch, 
				  404, 487, 412, 411, 414, 224,
				  # leotarrrd, pirate pelvis, safarrri hat, Snakehead Charrrm, stuffed shoulder parrot,
				  415, 406, 416, 485, 403,
				  # sunken chest, swashbuckling pants
				  405, 402],
		"ranch" => [142, 144], # anticheese, Stone of eXtreme Power
		"sewer" => [ # disco ball, disco mask, helmet turtle, mariachi pants, pasta spoon, ravioli hat, saucepan
			     10, 9, 3, 12, 5, 6, 7,
			     # scroll of turtle summoning, seal tooth, seal-clubbing club, stolen accordion
			     4, 2, 1, 11],
		"ski" => [145, 474], # Barbed-Wire Fence, cast
		"sleazybackalley" => [ # bum cheek, dirty hobo gloves, Mad Train wine, the Slug Lord's map
				       566, 565, 564, 598],
		"spookyforest" => [ # bar skin, barskin hat, dried face, helmet turtle, meat paste, mosquito larva, saucepan
				    70, 72, 98, 3, 25, 275, 7,
				    # scroll of turtle summoning, spooky sapling, spooky shrunken head, spooky stick,
				    4, 75, 102, 77,
				    # Spooky Temple map, wooden stakes
				    74, 71],
		"valley" => [ # 1337 7r0uZ0RZ, 30669 scroll, 33398 scroll, draggin' ball hat, Dyspepsi-Cola
			      542, 549, 550, 541, 347,
			      # f3d0r4, flaming talons, lowercase N, meat Vortex, pr0n legs, spam witch sammich 
			      538, 544, 539, 546, 537, 545,
			      # Tasty Good Fun Rice Candy, trollhouse cookies 
			      540, 543],
		"whiteysgrove" => [ # mullet wig, White Castle burger, White Castle fries, white chocolate chips,
				    267, 261, 260, 263,
				    # white lightning, white picket fence, white satin pants, white sword, whitesnake skin
				    266, 270, 265, 269, 259]);

# And set is the most clever verb of all (in some ways at least).

my %setobjs = ("debug" => \$debug, "verbose" => \$verbose, "noconnect" => \$noconnect,
	       "cloverprotect" => \$cloverprotect, "macrofail" => \$macrofail, "alwaysequip" => \$alwaysequip,
	       "questmax" => \$questmax, "summary" => \$summary, "maxdrunk" => \$maxdrunk, "fleewhen" => \$fleewhen,
	       "checkbuffbots" => \$checkbuffbots);
my %sethelp = ("debug" => "Turn on (1)/off (0) debug mode (prints out full HTML received)",
	       "verbose" => "Turn on (1)/off (0) verbose mose (not currently implemented)",
	       "noconnect" => "Turn on (1)/off (0) noconnect mode (doesn't actually connect)",
	       "cloverprotect" => "How carefully (0-3) to prevent adventuring when holding a clover",
	       "macrofail" => "Always stop (0)...never stop (3) macro execution on error",
	       "alwaysequip" => "If on (1), will always equip custom outfit before adventuring",
	       "questmax" => "Maximum number of adventures to spend questing",
	       "summary" => "Turn on (1)/off (0) display of adventure summary",
	       "maxdrunk" => "Maximum drunkenness (15 or 20) your character can tolerate",
	       "fleewhen" => "Flee from combat if your HP falls below this number",
	       "checkbuffbots" => "Check to see that the buffbots are online before submitting meat");
my %setvalues = ("debug" => [0,1], "verbose" => [0,1], "noconnect" => [0,1], "cloverprotect" => [0,1,2,3],
		 "macrofail" => [0,1,2,3], "alwaysequip" => [0,1], "summary" => [0,1], "maxdrunk" => [15,20],
		 "checkbuffbots" => [0,1]);

# Actions to take in combat. This will eventually be superceded by something better.
# For now, it's all just attacking, since this hash isn't actually implemented yet.

my %actions = ("attack" => "attack",
	       "moxious" => "moxman",
	       "indigestion" => "skill&whichskill=2",
	       "thrustsmack" => "skill&whichskill=1003",
	       "lungesmack" => "skill&whichskill=1004",
	       "ltsmack" => "skill&whichskill=1005",
	       "headbutt" => "skill&whichskill=2003",
	       "spectralsnapper" => "skill&whichskill=2005",
	       "minorray" => "skill&whichskill=3003",
# Entangling Noodles are disabled because they don't actually
# do any damage. So they're not going to win you a fight.
#	       "entangling" => "skill&whichskill=3004", 
	       "extremeray" => "skill&whichskill=3005",
	       "cone" => "skill&whichskill=3007",
	       "pastalord" => "skill&whichskill=3008",
	       "stream" => "skill&whichskill=4003",
	       "storm" => "skill&whichskill=4005",
	       "wave" => "skill&whichskill=4009",
	       "geyser" => "skill&whichskill=4012",
	       "eyepoke" => "skill&whichskill=5003",
	       "discodance" => "skill&whichskill=5005",
	       "ddance2" => "skill&whichskill=5008",
	       "facestab" => "skill&whichskill=5012");

my %actionverbs = ("attack" => "Attacking",
		   "moxious" => "Using a Moxious Maneuver",
		   "indigestion" => "Using Chronic Indigestion",
		   "thrustsmack" => "Using a Thrust-Smack",
		   "lungesmack" => "Using a Lunge-Smack",
		   "ltsmack" => "Using a Lunging Thrust-Smack",
		   "pastalord" => "Casting 'Weapon of the Pastalord'",
		   "eyepoke" => "Using Disco Eye-Poke",
		   "discodance" => "Using Disco Dance of Doom",
		   "ddance2" => "Using Disco Dance II: Electric Bugaloo",
		   "facestab" => "Using Disco Face-Stab"
		   );

my $action = $actions{"attack"};
my $attackverb = $actionverbs{"attack"};

# Pretty printer. Prints a list of things so that they don't have word wrap.
# First argument = leader, remaining list = list of things to be printed.

sub prettyprint($@) {
    my $width = 80;
    my $leader = shift;
    
    my $printline = join(", ", @_);
    my @words = split(" ", $printline);

    my $curpos = 0;
    if ($leader) {
	print $leader, ":";
	
	$curpos = length($leader) + 1;
    }
    
    while (@words) {
	my $word = shift @words;

	# If this word would take us past the edge of the screen,
	# AND we're not already at the left margin, then put in
	# a line break.

	if (((length($word) + $curpos) >= ($width-1)) &&
	    ($curpos != length($leader) + 1)) {
	    print "\n";
	    if ($leader) {
		$curpos = length($leader) + 1;
		print " " x $curpos;
	    }
	    else { $curpos = 0; }
	}

	print " ", $word;
	$curpos += length($word) + 1;
    }
    print ".\n";
}

# Adventure summary printer.

sub print_advsummary() {
    return if ($summary == 0);

    my $totitems;
    my $totgains;
    
    if (%advitems) { 
	foreach my $item (sort {lc($a) cmp lc($b)} keys %advitems) {
	    $totitems .= ", " if $totitems;
	    $totitems .= $item;
	    $totitems .= " ($advitems{$item})" if ($advitems{$item} > 1);
	}
	prettyprint "Total items", $totitems;
    }

    if (%advgains) {
	foreach my $gain (sort {lc($a) cmp lc($b)} keys %advgains) {
	    $totgains .= ", " if $totgains;
	    $totgains .= "$advgains{$gain} $gain";
	}
	prettyprint "Total gains", $totgains;
    }
}

# Moon/menu pane getter and parser. We call it "moonpane",
# because all we really care about are the moons.
#
# Information about moon semantics taken from KoL Coldfront:
# http://kol.coldfront.net/index.php/content/view/310/64/
#
# Thanks to phenyx for writing this code. I modified it
# to match my style somewhat (and made the output a little
# more concise).

sub get_moonpane($) {
    my $printresults = shift;
    my $bonus_day;
    my $bonus_tomorrow;
    my $grue = 0;
    my $wereseal = 0;
    my $baio = 0;
    my @moons;
    
    print "Getting moon status...\n";
    
    return 1 if $noconnect;
    
    # XXX I postulate that we can fetch compactmenu regardless
    # of whether we're in compact mode or not.
    # This saves us split parsing.
    &get_page("$site/compactmenu.php");

    @moons = ($response =~ m/\/itemimages\/smoon(\d+)\.gif/g);

    if (@moons != 2) {
	print "Error in getting moon status! Try again later.\n";
	return 0;
    }

    # Translate moon phases into verbose descriptions.
    my @moonphases = ("New", "Waxing 1/4", "Waxing 1/2", "Waxing 3/4",
		      "Full", "Waning 3/4", "Waning 1/2", "Waning 1/4");
    
    # Bonus day.
    if ($moons[0] == 5) { $bonus_day = "Mysticality"; }
    elsif ($moons[1] == 5) { $bonus_day = "Muscle"; }
    elsif ($moons[0] == $moons[1]) { $bonus_day = "Moxie"; }

    # Check to see if tomorrow is a bonus day.
    if ($moons[0] == 4) { $bonus_tomorrow = "Mysticality"; }
    elsif (($moons[0] == 8 && $moons[1] == 4) ||
	   ($moons[0] == 1 && $moons[1] == 5)) { $bonus_tomorrow = "Muscle"; }
    elsif (($moons[0] == 7 && $moons[1] == 8) ||
	   ($moons[0] == 8 && $moons[1] == 8)) { $bonus_tomorrow = "Moxie"; }

    # Calculate total fullness of moons (in quarters).
    my $totfullness = 0;
    foreach my $moon (@moons) {
	if ($moon <= 4) { $totfullness += $moon - 1; }
	else { $totfullness += 9 - $moon; }
    }

    # Figure out grue. Grue will fight iff the total fullness is <= 4 quarters.
    if ($totfullness <= 4) { $grue = 1; }
    
    # Figure out warsil.
    $wereseal = ($totfullness - 4) * 25;

    # Figure out Baio (+10% for each quarter of fullness)
    $baio = ($totfullness) * 10;
    
    if ($printresults) {
	print "Ronald: ", $moonphases[$moons[0]-1], ", Grimace: ", $moonphases[$moons[1]-1];
	print " ($bonus_day day today!)" if $bonus_day;
	print " ($bonus_tomorrow day tomorrow)" if $bonus_tomorrow;
	print "\n";
	print "(Wereseal: ", ($wereseal > 0) ? "+" : "", "$wereseal%; Baio: +$baio%; Grue: ",
	($grue) ? "will fight" : "won't fight", ")\n";
    }
    return 1;
}

# Charpane getter and parser. Give this a non-zero argument if you want it
# to print its results.

sub get_charpane($) {
    my $printresults = shift;
    my $meat;
    my $charname;
    my $charlevel;
    my $charclass = "";
    my $missed = 0;
    my $compactmode = 0;
    
    my ($mus, $mys, $mox, $modmus, $modmys, $modmox);

    print "Getting character status...\n";

    return 1 if $noconnect;

    &get_page("$site/charpane.php");

    # Most of the parsing is in two passes, to deal with compact mode.
    # The first pass is for regular mode; if it misses, the second
    # searches for compact mode.

    # Character name
    if ($response =~ /charsheet\.php\"><b>(.*?)<\/b>(?:.*?)>Level (\d+)<br>([^<>]+?)</) {
        $charname = $1;
	$charlevel = $2;
	$charclass = $3;
	if ($charname !~ /$username/i) { print "Warning: Character name in character pane $1 does not match login name $username!\n"; }
    }
    elsif ($response =~ /charsheet\.php\">(.*?)<\/a>(?:.*?)>Lvl\. (\d+)</) {
	$compactmode = 1;
	$charname = $1;
	$charlevel = $2;
	if ($charname !~ /$username/i) { print "Warning: Character name in character pane $1 does not match login name $username!\n"; }
    }
    else { $missed = 1; }

    # Muscle, mysticality, moxie. These patterns are a little tricky
    # to understand. The trick is to realize that the ^: prevents
    # us from falling through to the next stat.
    
    if ($response =~ /Mus(?:cle)?:[^:]*?(\d+)[^:]*?\((\d+)\)/) {
	$modmus = $1;
	$mus = $2;
    }
    elsif ($response =~ /Mus(?:cle)?:[^:]*?(\d+)/) {
	$modmus = $1;
	$mus = $1;
    }
    else { $ missed = 1; }

    if ($response =~ /Mys(?:ticality)?:[^:]*?(\d+)[^:]*?\((\d+)\)/) {
	$modmys = $1;
	$mys = $2;
    }
    elsif ($response =~ /Mys(?:ticality)?:[^:]*?(\d+)/) {
	$modmys = $1;
	$mys = $1;
    }
    else { $ missed = 1; }

    if ($response =~ /Mox(?:ie)?:[^:]*?(\d+)[^:]*?\((\d+)\)/) {
	$modmox = $1;
	$mox = $2;
    }
    elsif ($response =~ /Mox(?:ie)?:[^:]*?(\d+)/) {
	$modmox = $1;
	$mox = $1;
    }
    else { $ missed = 1; }

    # HP
    if ($response =~ /hp\.gif.*?>(\d+)&nbsp;\/&nbsp;(\d+)</) {
	$curhp=$1;
	$maxhp=$2;
    }
    elsif ($response =~ /HP.*?>(\d+)\/(\d+)</) {
	$curhp=$1;
	$maxhp=$2;
    }
    else { $missed = 1; }

    # MP
    if ($response =~ /mp\.gif.*?>(\d+)&nbsp;\/&nbsp;(\d+)</) {
	$curmp=$1;
	$maxmp=$2;
    }
    elsif ($response =~ /MP.*?>(\d+)\/(\d+)</) {
	$curmp=$1;
	$maxmp=$2;
    }
    else { $missed = 1;}

    # Meat. This has a special catch to deal with commas.
    # I don't think any of the other quantities have commas
    # in them, but if they do I'll need to adjust appropriately.
    if ($response =~ /meat\.gif.*?>([0-9,]+)</) {
	$meat = $1;
    }
    elsif ($response =~ /Meat.*?>([0-9,]+)</) {
	$meat = $1;
    }
    else { $missed = 1; }

    # Adventures
    if ($response =~ /hourglass\.gif.*?>(\d+)</) {
	$advleft = $1;
    }
    elsif ($response =~ /Adv.*?>(\d+)</) {
	$advleft = $1;
    }
    else { $missed = 1; }

    # Drunkenness
    $drunk = 0;
    if ($response =~ /(?:Inebriety|Tipsiness|Drunkenness|Temulency|Drunk).*?>(\d+)</) {
	$drunk = $1;
    }
    # Don't set the missed flag if we don't get drunkenness -- maybe
    # we're not drunk.

    if ($missed == 1) {
	print "Error: Parsing failed in get_charpane\n";
	
	# Kill advleft so we don't go off adventuring too much.
	$advleft = 0;

	# Set the error flag. This isn't a fatal error (unless called
	# from within "until", in which case the main parser loop will
	# take care of it).
	
	$macroerror = 1 if ($macroerror < 1);
	return 0;
    }
    
    if ($printresults) {
	print "Character status:\n";
	print "$charname: Level $charlevel $charclass\n";
	if ($modmus == $mus) { print "$mus Mus, "; }
	else { print "$modmus ($mus) Mus, "; }
	if ($modmys == $mys) { print "$mys Mys, "; }
	else { print "$modmys ($mys) Mys, "; }
	if ($modmox == $mox) { print "$mox Mox\n"; }
	else { print "$modmox ($mox) Mox\n"; }
	print "$curhp/$maxhp HP, $curmp/$maxmp MP, $meat Meat, $advleft Adv, $drunk Drunk\n";
    }
    
    $hassphere=0;
    my $effects;

    if ($compactmode == 0) {
	while ($response =~ /desc_effect.php.*?>([^<>]+?)</g) {
	    my $cureffect = $1;
	    $effects .= ", " if $effects;
	    $effects .= $cureffect;
	    if ($cureffect =~ /Elemental Saucesphere/) { $hassphere=1; }
	}
    }
    else {
	while ($response =~ /alt=\"(.*?)\".*?desc_effect\.php.*?>([^<>]+?)</g) {
	    my $cureffect = "$1 $2";
	    $effects .= ", " if $effects;
	    $effects .= $cureffect;
	    if ($cureffect =~ /Elemental Saucesphere/) { $hassphere=1; }
	}
    }

    prettyprint("Effects", $effects) if ($effects && $printresults);

    # In compact mode, we only get the familiar weight. There's nothing
    # we can do about this.

    if ($response =~ /familiar\.php.*?>([^<>]+?)<(?:.*)?the (\d+) pound ([^<>]+?)</) {
	$famname = $1;
	$famweight = $2;
	$famtype = $3;
	print "Familiar: $1, the $famweight pound $famtype\n" if $printresults;
    }
    elsif ($response =~ /familiar\.php.*?>(?:.*?)familiar(\d+)\.gif(?:.*?)<br>(\d+) lbs?\.</) {
	if ($familiarnames{$1}) {
	    print "Familiar: a $2 pound $familiarnames{$1}\n" if $printresults;
	}
	else {
	    print "Familiar: $2 pounds\n" if $printresults;
	}
    }
    return 1;
}

# List familiars function. This parses the familiar page and prints
# out the name, type, weight, and number of kills of each unequipped
# familiar.  It also prints the name and type of the equipped familiar,
# the total number of familiars, and total familiar weight.
# It is very similar to the actual familiar page.  
# Give this a non-zero argument if you want it to print its results.
# Written by nithan (166143).
# Code somewhat altered by me.

sub get_familiarlist($) {
    my $printresults = shift;
    my $terrarium;
    my $totalfams=0;
    my $totalweight=0;
    my $totaltrophy=0;
    
    print "Getting familiar list...\n" if $printresults;
    
    return 1 if $noconnect;
    
    &get_page("$site/familiar.php");
    
    # Familiar equipped?
    if ($response =~ /Current Familiar(?:.*?)<b>(.*?)<\/b><br>(\d+) pound ([^<>]+?)</) {
	$totalfams++;
	$totalweight += $2;
	$totaltrophy += $2;
	print "Current familiar: $1, the $2-pound $3\n" if $printresults;
    } else {
	print "You do not currently have a familiar with you.\n" if $printresults;
    }
    
    # The Current Familiar html doesn't fit the patterns below, so kill it;
    # $terrarium contains all remaining familiars 
    $terrarium = $response;
    $terrarium =~ s/(.*?)creatures//;
    
    # Loop through the terrarium 
    while ($terrarium =~ /desc_familiar\.php(.*?)<\/tr>/g) {
	my $fam=$1;
	# Check to see if the familiar has an accessory.
	if ($fam =~ /<b>(.*?)<\/b>, the (\d+) pound (.*?) \((\d+)(?:.*?)\/itemimages\/(.*?)\.gif/g) {
	    $totalfams++;
	    $totalweight += $2;
	    # Familiar weights only count towards the trophy unless they have at
	    # least 1 kill.
	    $totaltrophy += $2 if $4;
	    print "$1, the $2-pound $3 ($4 kills) ($5)\n" if $printresults;
	} elsif ($fam =~ /<b>(.*?)<\/b>, the (\d+) pound (.*?) \((\d+)(.*?)/g) {
	    # no familiar accessory
	    $totalfams++;
	    $totalweight += $2;
	    $totaltrophy += $2 if $4;
	    print "$1, the $2-pound $3 ($4 kills)\n" if $printresults;
	} else {
	    print "Error parsing familiar.\n";
	}
	
    } # end while
    print "You have $totalfams familiars weighing $totalweight pounds";
    print " ($totaltrophy for trophy purposes)" if ($totalweight != $totaltrophy);
    print ".\n";
    return 1;
}

# Start function. This sets up the cookie file, gets the password
# and password hash, and performs login. This is in its own routine
# so that switch can just call it.

sub start_new_user() {
    # Forget any custom outfits.
    %customoutfits = ();
    $gotoutfits = 0;

    $options = "-L $ua $cookies --silent ";
    
    # Don't put strange characters into the cookie file name. This
    # could allow all kinds of mischief.

    $cookiefilename = $username;
    $cookiefilename =~ tr/A-Za-z0-9_\- //cd;
    $cookiefilename =~ s/ /_/g;
    $options =~ s/USERNAME/$cookiefilename/g;

    if ($passwords{$username}) {
	$password = $passwords{$username};
    }
    else {
	print "Password for $username: ";
	$password = <STDIN>;
	chomp $password;
    }

    &do_login();

    # Set the settings from the config file. If you can understand this,
    # your Perl-fu is mighty indeed.

    if ($settings{$username}) {
	foreach my $param (keys % {$settings{$username}}) {
	    $ {$setobjs{$param}} = $settings{$username}{$param};
	}
    }

    &get_charpane(1) or die "Error in processing character pane. Try logging in again.\n";
    # The result of the moonpane is not so critical.
    &get_moonpane(1);
}

# Login subroutine.

sub do_login() {
    return 1 if $noconnect;

    # Check to make sure we haven't exceeded our relogin limit. See get_page() for
    # more details.

    if ($relogins > 5) {
	die "\n\nRelogin limit exceeded. Check to make sure there isn't something wrong.\n";
    }

    print "Logging in $username...";

    # Do some quick-and-dirty URL encoding. AFAIK, this is the only string
    # that we send to the server that will require such operations (except for
    # names in buffs, which we also handle).

    my $codedname = $username;
    my $codedpw = $password;
    $codedname =~ s/([^-A-Za-z0-9_])/sprintf("%%%02X", ord($1))/ge;
    $codedpw =~ s/([^-A-Za-z0-9_])/sprintf("%%%02X", ord($1))/ge;

    my $paramstring = "loggingin=Yup.&loginname=$codedname&password=$codedpw";

    my @response = `curl $options -d "$paramstring" $site/login.php`;
    my $response = join("", @response);

    # Failing to login is a fatal error. We should die here since do_login will be called
    # either from the main routine or as a relogin.

    if ($response =~ /down for nightly maintenance/) {
	die "\n\nKoL is down for nightly maintenance. Try again later.\n";
    }
    if ($response =~ /Can\'t connect to MySQL server/) {
	die "\n\nKoL seems to be having some SQL problems. Try again later.\n";
    }
    if ($response =~ /Login failed/) {
	die "\n\nSorry, incorrect login name or password.\n";
    }
    if ($response =~ /<b>Invalid/i) {
	die "\n\nInvalid Session ID. Try logging in again.\n";
    }
    if ($response !~ /main\.php/) {
	print "\n\nUnknown error in login. Got this page from the server:\n";
	die $response, "\n";
    }
    
    print "   ....Logged in!\n";

    $currentoutfit = -99;  # If we're relogging in, our outfit might have changed.
    
    # Since the password hash now changes across sessions, we
    # have to get it anew each time.

    $pwdhash = &get_pwdhash() or die "Couldn't get password hash. Try logging in again.\n";
    
    return 1;
}

# Get password hash. If the password hash is not specified in the config file
# go to skills.php and obtain the password hash.

sub get_pwdhash() {
    print "Getting password hash...\n";
    
    if ($noconnect == 1) {
	print "curl $options -d whichstore=m $site/store.php";
	return 1;
    }
    
    my @response = `curl $options -d whichstore=m $site/store.php`;
    my $response = join("", @response);
    
    # Failing to get the password hash is a fatal error, but
    # since get_pwdhash is only called once, we'll let the main
    # routine handle it.
    
    if ($response =~ /name=phash value=[\'\"]?([A-Za-z0-9]*)[\'\"]?/) {
	return $1;
    }
    else { return 0; }
}

# get_page. This is essentially a wrapper around calls to curl
# which will check to see that we haven't been hosed by timeout
# or invalid session ID and, if so, relogin. It also dies on
# some fatal errors (maintenance, etc.)

# Note that get_page always either succeeds or dies, so there's no
# need to propagate errors out of here.

sub get_page($) {
    my $params = shift;
    
    if ($noconnect == 1) {
	print "curl $options $params\n";
	return 1;
    }

    my $success = 0;
    my $retries = 0; # Retries for no response from server...not to be
                     # confused with relogins, which have their own
                     # counter persisting across the entire session

    while ($success == 0) {
	
	# We have to do the password hash substitution here now,
	# since if we need to relogin, the password hash will change.

	my $hashedparams = $params;
	$hashedparams =~ s/PASSWORDHASH/$pwdhash/g;

	@response = `curl $options $hashedparams`;
	$response = join(" ", @response);

	# Eliminate tildes and other such characters, since they don't display so well in ASCII.
	$response =~ s/&ntilde;/n/g;
	$response =~ s/&quot;/\"/g;

	print $response if $debug;
	
	# Check for fatal errors.
	
	if ($response =~ /down for nightly maintenance/) {
	    die "\n\nKoL is down for nightly maintenance. Try again later.\n";
	}
	if ($response =~ /Can\'t connect to MySQL server/) {
	    die "\n\nKoL seems to be having some SQL problems. Try again later.\n";
	}
	
	# Check for nonfatal errors. If we've timed out or gotten an invalid session
	# ID, then set success to 0 (so we try the request again) and call do_login.
	# do_login will die if it fails, but we should also have a maximum number
	# of retries so that pathological situations like two competing sessions
	# don't cause an infinite loop.
	
	$success = 1;
	if ($response =~ /<b>Invalid/i) {
	    $success = 0;
	    $relogins++;
	    print "Invalid Session ID. Attempting to relogin...\n";
	    &do_login();
	}
	elsif ($response =~ /input type=hidden name=loggingin value=\"Yup\.\"/) {
	    $success = 0;
	    $relogins++;
	    print "Session timed out. Attempting to relogin...\n";
	    &do_login();
	}
	elsif ($response =~ /^\s*$/) {
	    $success = 0;
	    if ($retries > 5) {
		die "Got too many blank responses from server.\n";
	    }
	    print "Got blank response from server. Trying again...\n";
	    $retries++;
	}
    }
}
    
# Do some basic parsing of the response to figure out what we got.
# Right now this only looks for item gains and stat gains. It works
# with adventures and non-adventures, but if you want any more parsing
# beyond item gains and stat gains, you'll have to do it in the
# calling routine.

sub parse_response() {
    my $muswords = "Beefiness|Fortitude|Muscleboundness|Strengthliness|Strongness";
    my $myswords = "Enchantedness|Magicalness|Mysteriousness|Wizardliness";
    my $moxwords = "Cheek|Chutzpah|Roguishness|Sarcasm|Smarm";

    my $items;
    while ($response =~ m/You acquire (?:an item|some items): <b>(.*?)<\/b>/g) {
	my $stuff = $1;
	my $amount = 1;

	$items .= ", " if $items;
	$items .= $stuff;

	if ($stuff =~ /^(.*) \((\d+)\)$/) {
	    $stuff = $1;
	    $amount = $2;
	}

	$n += $amount if ($stuff =~ /worthless/);
	$questcur += $amount if ($onquest == 1 && $stuff =~ /\Q$questobj\E/i);
	$advitems{$stuff} += $amount;
    }

    # Thanks to phenyx for writing the Meat parser. He also did the 
    # muscle/mysticality/moxie points, too, but I had already beaten
    # him to it, so the code here is mine. (It's pretty much the same
    # though.)
    
    my $gains;
    while ($response =~ m/You gain ([^<]*?)[.!]</g) {
	my $what = $1;
	
	if ($what =~ /(\d+) (.*)/) {
	    my $qty = $1;
	    my $gained = $2;

	    if ($gained eq "Meat") { $advgains{$gained} += $qty; }
	    elsif ($gained =~ /$muswords/o) { $advgains{"muscle"} += $qty; }
	    elsif ($gained =~ /$myswords/o) { $advgains{"mysticality"} += $qty; }
	    elsif ($gained =~ /$moxwords/o) { $advgains{"moxie"} += $qty; }
	    elsif ($gained eq "Drunkenness") { $drunk += $qty; }
	    elsif ($gained =~ /hit point/) {
		$curhp += $qty;
		$curhp = $maxhp if ($curhp > $maxhp);
	    }
	}

	$gains .= ", " if $gains;
	$gains .= $what;
    }
    
    my $losses;
    while ($response =~ m/You lose ([^<]*?)[.!]</g) {
	my $what = $1;

	if ($what =~ /(\d+) (.*)/) {
	    my $qty = $1;
	    my $lost = $2;

	    if ($lost eq "Meat") { $advgains{$lost} -= $qty; }
	    elsif ($lost =~ /$muswords/o) { $advgains{"muscle"} -= $qty; }
	    elsif ($lost =~ /$myswords/o) { $advgains{"mysticality"} -= $qty; }
	    elsif ($lost =~ /$moxwords/o) { $advgains{"moxie"} -= $qty; }
	    elsif ($lost eq "Drunkenness") { $drunk -= $qty; }
	    elsif ($lost =~ /hit point/) {
		$curhp -= $qty;
		$curhp = 0 if ($curhp < 0);
	    }
	}

	$losses .= ", " if $losses;
	$losses .= $what;
    }

    my $effects;
    while ($response =~ m/You acquire an effect: <b>(.*?)<\/b><br>\(duration: (\d+) Adventures\)/ig) {
	$effects .= ", " if $effects;
	$effects .= "$1 ($2 turns)";
    }

    if ($items) { prettyprint "        You got items", $items; }
    if ($gains) { print "\tYou gain $gains.\n"; }
    if ($losses) { print "\tYou lose $losses.\n"; }
    if ($effects) { prettyprint "        You acquire effects", $effects; }
}

# Battle parser. Give it a 1 if this is the first time you're calling
# it in a given battle and 0 otherwise.

sub parse_battle_results($) {
    my $firstcall = shift;
    my $monster;
    my $win = 0;

    if ($response =~ /You\'re fighting (.*?)<\/td>/) { $monster = $1; }
    else { $monster = "Error: Couldn't extract monster name\n"; }

    if ($response =~ /<!--WINWINWIN-->/) { $win = 1; }

    # Display what we're fighting, but only once!
    if ($firstcall == 1) {
	print "\tYou're fighting $monster...\n";
    }

    # Look for damage done to and by us.
    if ($response =~ /(You hit for \d+ damage\.)/) {
	print "\t", $1;
	if ($response =~ /CRITICAL HIT/) { print " CRITICAL HIT!"; }
	print "\n";
    }
    elsif ($response =~ /(You failed to hit the monster\.)/) {
	print "\t", $1, "\n";
    }
    elsif ($response =~ /FUMBLE/ && $response =~ /(You drop[^.]*? (\d+) damage\.)/) {
	print "\tFUMBLE! $1\n";
	$curhp -= $2;
	$curhp = 0 if ($curhp < 0);
    }
    elsif ($response =~ /(You attempt to run away like a sissy little coward, but you don\'t quite make it\.)/) {
	print "\t$1\n";
    }
    elsif ($response !~ /gets the jump on you/) { print "Couldn't figure out whether you hit or missed.\n"; }

    if ($win==0) {
	# The enemy only acts if we don't kill him, so we should only
	# check for something in that case.
	if ($response =~ /((?:He|She|It|They) tr(?:y|ies)[^<]*miss(?:es)?\.)/) {
	    print "\t", $1, "\n";
	}
	elsif ($response =~ /((?:He|She|It|They)[^<]*\.)<b>/) {
	    print "\t", $1, "\n";
	}
	elsif ($response =~ /((?:He|She|It|They) tr(?:y|ies)[^<]*)</) {
	    print "\t", $1, "\n";
	}
	elsif ($firstcall==0) {
	    # If we strike first, sometimes the enemy doesn't get to strike back.
	    # But sometimes they do.
	    print "Couldn't figure out whether the enemy hit or missed.\n"; 
	    print $response, "\n";
	}
    }

    # See what our familiar is up to. If we have one, that is.

    if (defined $famname) { # Just in case some jerk has a familiar named "0"
	if ($response =~ /(\Q$famname\E[^<]*?\.)/) {
	    print "\t$1\n";
	}
	if ($response =~ /(\Q$famname\E[^.]*?\!)/) {
	    print "\t$1\n";
	}    
    }

    # See what the Detective Skull says.

    if ($response =~ /Your Detective Skull says \"I deduce that this monster has approximately (\d+) hit points.\"/) {
	print "\tAccording to your Detective Skull, the monster has about $1 HP left.\n";
    }

    # If we won, say so. Otherwise, call parse_response to display
    # our gains and losses (HP/MP from enemies and/or familiars).

    if ($win) {	print "\tYou win the fight!\n"; }
    else { &parse_response(); }
}

# Equip a familiar. Feed this subroutine a familiar name.
# "familiar none" functionality added by nithan.

sub equip_familiar($) {
    my $familiarname = shift;

    if ($familiarname eq "list") {
	return &get_familiarlist(1);
    }

    if (!$familiarobjs{$familiarname}) {
	print "Error in equip_familiar: Unknown familiar $familiarname\n";
	$macroerror = 3 if ($macroerror < 3);
	return 0;
    }
    my $familiar = $familiarobjs{$familiarname};

    print "Changing familiar... ";
    if ($familiar == -1) {
	&get_page("-d \"action=putback\" $site/familiar.php");
	if ($response =~ /(You put .*? back in the Terrarium\.)/) {
	    print $1, "\n";
	    return 1;
	}
	else {
	    print "You do not currently have a familiar with you.";
	    return 0;
	}
    } else {

	&get_page("-d \"action=newfam&newfam=$familiar\" $site/familiar.php");

	if ($response =~ /(You take (.*?) with you\.)/) {
	    print $1, "\n";
	    $famname = $2;
	    return 1;
	}
	else { 
	    # This is tricky. It might have failed because we are already carrying that
	    # familiar, or it might have failed because we don't have that familiar.
	    # Distinguishing between the two cases requires looking at the current
	    # familiar (if indeed we *have* one).
	    
	    if ($response =~ /(Current Familiar.*?)form/) {
		my $currentfam = $1;
		if ($currentfam =~ /$familiarname/i) {
		    print "That familiar is already with you!\n";
		    return 1;
		}
	    }
	    if ($response =~ /You don\'t have a familiar of that type/) {
		print "You don't have that familiar.\n";
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    }
	}
	print "Unexpected error in equip_familiar. Got this from the server:\n";
	print $response, "\n";
	$macroerror = 3 if ($macroerror < 3);
    }
}

# Routine to toggle between full and compact mode.
# Written by nithan.

sub mode_toggle() {
    my $printresults = shift;
    
    print "Toggling mode... ";
    &get_page("-d \"action=togglecompact\" $site/account.php");
    
    if ($response =~ /Switch to Compact Mode/) {
	print "You have switched to Full mode.\n";
	return 1;
    }
    elsif ($response =~ /Switch back to Full Mode/) {
	print "You have switched to Compact mode.\n";
	return 1;
    } else {
	print "Unexpected response while switching modes.\n";
	return 0;
    }
}

# Equip an outfit. Feed this subroutine an outfit name, not a number.

sub equip_outfit($) {
    my $outfitname = shift;
    
    # If we don't know what outfit this is, then we should
    # check the server to see if it's a custom outfit.

    if (!$outfitobjs{$outfitname} && $gotoutfits == 0) {
	# Get the list of custom outfits from the server.
	print "Getting custom outfit list...";
	
	&get_page("$site/inventory.php?which=2");

	while ($response =~ /<option value=([^<>]+?)>Custom: (.*?)</g) {
	    my $outfitname = $2;
	    my $outfitnum = $1;
	    $customoutfits{$outfitname} = $outfitnum;
	}
	$gotoutfits = 1;
	print "Done.\n";
    }
    
    # Now check to see if it's a custom outfit. The
    # parsing wasn't done in the main loop, so we have
    # to do it ourselves. Bummer.

    my @potentials = grep (/^\Q$outfitname/i, (keys %outfitobjs, keys %customoutfits));
    if (@potentials == 0) {
	print "Error in equip_outfit: Unknown outfit $outfitname\n";
	$macroerror = 3 if ($macroerror < 3);
	return 0;
    }
    elsif (@potentials > 1) {
	print "Error in equip_outfit: Ambiguous outfit: ", join(", ", @potentials), "\n";
	$macroerror = 3 if ($macroerror < 3);
	return 0;
    }

    my $outfit = $outfitobjs{$potentials[0]} || $customoutfits{$potentials[0]};
    
    if ($outfit == $currentoutfit) {
	print "Already equipped with the right outfit!\n";
	return 1;
    }

    print "Equipping outfit... ";
    &get_page("-d \"action=outfit&which=2&whichoutfit=$outfit\" $site/inv_equip.php");
    
    # If we don't yet have the list of outfits, we can now get
    # it from here for free. (Even if the equip didn't succeed.) Yay!

    if ($gotoutfits == 0) {
	while ($response =~ /<option value=([^<>]+?)>Custom: (.*?)</g) {
	    my $outfitname = $2;
	    my $outfitnum = $1;
	    $customoutfits{$outfitname} = $outfitnum;
	}
	$gotoutfits = 1;
    }

    if ($response =~ /<td>You put on an Outfit: <b>(.*?)<\/b>/) {
	if (!$1) { print "Equipped Custom Outfit.\n"; }
	else { print "Equipped $1.\n"; }
	$currentoutfit = $outfit;
	return 1;
    }
    else {
	print "Error equipping outfit. Perhaps you don't have the parts.\n";
	$macroerror = 2 if ($macroerror < 2);
	return 0;
    }
}

# Trinket-getting routine. This is pretty straightforward.
# One argument: number of trinkets to get.

sub get_trinket($) {
    my $ntrinkets = shift;
    my $trinketsgot = 0;

    while ($trinketsgot < $ntrinkets) {
	my $success = do_action($verbs{"buy"}, $buyobjs{"gum"}, $ntrinkets - $trinketsgot);
	$success &&= adventure("sewer", $ntrinkets - $trinketsgot);
	$trinketsgot += $n;
	if ($success == 0) {
	    # Failed at one point in the procedure. Set $n appropriately
	    # and get out.
	    $n = $trinketsgot;
	    $macroerror = 1 if ($macroerror < 1);
	    return 0;
	}
    }
    $n = $trinketsgot;
    return $n;
}

# Sell junk.
# This iterates over all items in a predefined list and sells them all.

sub sell_junk($) {
    my $area = shift;
    if (!$junkobjs{$area}) {
	print "Error in sell_junk: Unknown area $area\n";
    }
    my @junkitems = @{ $junkobjs{$area} };

    foreach my $junkitem (@junkitems) {
	&get_page("-d \"action=sell&pwd=PASSWORDHASH&whichitem=$junkitem&type=all\" $site/sellstuff.php");
	if ($response =~ /(You sell your.*?\.)/) { print $1, "\n"; }
    }
}

# Visit the hermit.
# Hermit visiting is a little complicated because the item that the hermit wants in exchange
# varies from visit to visit. So, we have to visit hermit.php once first, extract the 
# information from the page, and then perform the transactions. Fortunately, we can take
# advantage of the fact that for multiple exchanges, we can use the response from previous
# exchanges to get information for the next one.
# Takes two parameters: item name to get and number of visits.

sub visit_hermit($$) {
    my $getitem = shift;
    my $visits = shift;

    if (!$hermititems{$getitem}) {
	print "Error in visit_hermit: Unknown item $getitem\n";
	$macroerror = 3 if ($macroerror < 3);
	return 0;
    }

    &get_page("$site/hermit.php");

    for (my $i=1; $i<=$visits; $i++) {
	print "  Hermit visit $i / $visits...\n";
	
	if ($response =~ /sends you packing/) {
	    print "\tYou don't have anything the hermit wants.\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}

	if ($response =~ /You don\'t have a Hermit Permit/) {
	    print "\tYou don't have a Hermit Permit.\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}
	    
	if ($response !~ /It seems he wants to trade/) {
	    print "Unexpected response: didn't get what appears to be the hermit page.\n";
	    $macroerror = 3 if ($macroerror < 3);
	    return 0;
	}

	if ($response =~ /name=hermitwants value=(\d+)/) {
	    my $hermitwants=$1;
	
	    &get_page("-d \"action=Yep.&pwd=$pwdhash&hermitwants=$hermitwants&whichitem=$hermititems{$getitem}\" $site/hermit.php");

	    if ($response =~ /(You trade .*?)</) {
		print "\t$1\n";
	    } elsif ($response =~ /doesn\'t have that item/) {
		print "\tThe Hermit doesn't have that item.\n";
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    } else {
		print "Unexpected result from the hermit:\n";
		print $response, "\n";
		$macroerror = 3 if ($macroerror < 3);
		return 0;
	    }

	    &parse_response();
    
	}
	else {
	    print "Error in visit_hermit: Couldn't get hermitwants\n";
	    $macroerror = 3 if ($macroerror < 3);
	    return 0;
	}
    } # visit loop
    
    return 1;
}

# Non-adventuring subroutine. Takes three arguments: page to visit,
# parameters to use at that page, and number of turns.

sub do_action($$$) {
    my $actionpage = shift;
    my $actionparam = shift;
    my $num = shift;
    my $nturns = $num;
    my $suppressparse;   # a little hack for mallternatives

    %advitems = ();
    %advgains = ();

    # If redeeming tickets, dress appropriately.

    if ($actionparam =~ /whichitem=400/) {
	unless (&equip_outfit("fratboy")) {
	    print "Couldn't equip outfit for redeeming tickets.\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}
    }

    # If we're buying things, it makes sense to buy all at once rather than just buy
    # one each time. Similarly for gym visits, donations, and skills.

    if ($actionparam =~ /QTY/) {
	$actionparam =~ s/QTY/$nturns/;
	$nturns=1;
    }

    for (my $i=1;$i<=$nturns;$i++) {
	print "  Action $i / $nturns\n";
	
	&get_page("-d \"$actionparam\" $site/$actionpage");

	# Right now, the parser (such as it is) will take the first clump of HTML, after
	# the "Results:" box, whatever it may be. Then, it checks for error conditions and
	# returns an error if that happens. Hopefully nothing will fall through the cracks
	# this way.
	
	if ($response =~ /<blockquote>(.+?)</ ||
	    $response =~ /<center><table><tr><td>(?:<.+?> *)*(.+?)</) {
	    my $result = $1;
	    $result =~ s/[:,] *$/./;
	    print "\t$result\n" if $result;
	}
	
	# Mall fallback if we run into a limit.

	if ($response =~ /You may only buy/ && $mallobjs{"2$object"}) {
	    # Note that this will change $response so that the test below
	    # will operate on the new $response, which is what we want.
	    print "\tAttempting to buy from alternate vendor...\n";
	    $object = "2$object";
	    &do_action($verbs{"mallbuy"}, $mallobjs{"$object"}, $num);

	    # The above call to do_action will already parse the response,
	    # so we don't need to do it again.
	    $suppressparse=1;
	}

	# Special catch for clan failure.
	if ($response =~  /<b>Apply to a Clan/i) {
	    print "\tYou're not in a clan.\n";
	}
	
	# Check to see if our action failed. If so, don't keep trying.

	if ($response =~ /You don\'t have enough Adventures/ || 
	    $response =~ /You don\'t have that many adventures/ ||
	    $response =~ /You don\'t have any adventures/ ||
	    $response =~ /You can\'t afford/ ||
	    $response =~ /You don\'t have that much meat/ ||
	    $response =~ /You don\'t have that much Meat/ ||
	    $response =~ /You don\'t have enough meat/ ||
	    $response =~ /You don\'t have the item/ ||
	    $response =~ /You don\'t have that many/ ||
	    $response =~ /You\'re too drunk/ ||
	    $response =~ /spleen/ ||
	    $response =~ /too full/ ||
	    $response =~ /way too drunk/ ||
	    $response =~ /out of toast/ ||
	    $response =~ /You don\'t have a toaster/ ||
	    $response =~ /You don\'t have any Evil/ ||
	    $response =~ /You can only order food from the Arches once per day/ ||
	    $response =~ /You can only make/ ||
	    $response =~ /You may only buy/ ||
	    $response =~ /This store doesn\'t have that item at that price/ ||
	    $response =~ /That\'s not enough Meat/ ||
	    $response =~ /You don\'t have that key/ ||
	    $response =~ /You cannot donate that much/ ||
	    $response =~ /Sorry, kid/ ||
	    $response =~ /<b>Apply to a Clan/ ||
	    $response =~ /pants on, er, the heater/ ||
	    $response =~ /Nobody in your clan/ ||
	    $response =~ /Your coffer doesn\'t contain enough Meat/ ||
	    $response =~ /You don\'t have enough Mana/ ||
	    $response =~ /Invalid target player/) {
	    $macroerror = 1 if ($macroerror < 1);

	    # But still print out the summary if there's anything to summarize.
	    if ($actionpage eq "shore.php" && $nturns > 1) {
		print_advsummary();
	    }
	    return 0;
	}

	&parse_response() unless $suppressparse;

	# Check for bartender explosion.
	if ($response =~ /It begins to vibrate noisily, spits out a few drinks at random, and then explodes/) {
	    print "Your bartender-in-a-box blew up!\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}
    }

    # Print out the total items gained if we went to the shore.
    
    if ($actionpage eq "shore.php" && $nturns > 1) {
	print_advsummary();
    }

    return 1;
}

# Adventuring subroutine. Takes two arguments at the moment:
# area name and number of visits.

sub adventure($$) {
    my $areaname = shift;
    my $nturns = shift;
    
    %advitems = ();
    %advgains = ();

    # Reset counter if we're in the sewer.
    if ($areaname eq "sewer") { $n = 0; }

    if (!$areas{$areaname}) {
	print "  Error in adventure: unknown area $areaname\n";
	$macroerror = 3 if ($macroerror < 3);
	return 0;
    }
    my $areanum = $areas{$areaname};

    for (my $i=1;$i<=$nturns;$i++) {
	
	# Check our drunkenness. This goes inside the loop because certain
	# adventures may increase our drunkenness (e.g. Typical Tavern).
	# Characters without Liver of Steel are still responsible for setting
	# maxdrunk appropriately.

	if ($drunk >= $maxdrunk && $areaname ne "stupor") {
	    print "  You're too drunk to adventure!\n";
	    $macroerror = 1 if ($macroerror < 1);
	    return 0;
	}

	# Pick a random area.
	if ($areaname eq "random" || $areaname eq "arandom") {
	    $areanum = -1;
	    while ($areanum == -1 || ($areaname eq "random" &&
				      ($areanum == 25 || $areanum == 67 || $areanum == 29 ||  # might get drunk here
				       $areanum == 65 || $areanum == 36 ||                    # these require disguises
				       $areanum == 37 ||                                      # pickle factory
				       $areanum == 12 ||                                      # sewer
				       $areanum < 0   ||                                      # don't pick another random area!
				       $areanum == 73 || $areanum == 78))) {                  # require special equipment
		my @areapossibilities = values (%areas);
		my $num = int(rand(($#areapossibilities + 1)));
		$areanum = $areapossibilities[$num];
	    }
	}

	# These next steps only need to be done at the beginning,
	# unless we're doing random, in which case they need to be
	# done for each adventure.

	if ($i == 1 || $areaname eq "random" || $areaname eq "arandom") {

	    # Implementation of our patented CloverProtect(tm) technology!
	    # (Note: Technology is neither patented nor trademarked.)
	    if ($areanum == 12 ||                  # never go into the sewer with a clover
		# clover protection level 1: knob outskirts, knob treasury, haiku dungeon,
		# limerick dungeon, haunted pantry, hidden temple
		($cloverprotect >=1 && ($areanum==11 || $areanum==41 || $areanum==16 ||
					$areanum==19 || $areanum==10 || $areanum==17)) ||
		# clover protection level 2: add knob kitchen and frat house (in disguise)
		($cloverprotect >=2 && ($areanum==40 || $areanum==29)) ||
		# clover protection level 3: all areas
		$cloverprotect==3) {

		print "  Checking clover possession...\n";
		&get_page("-d which=3 $site/inventory.php");
		
		if ($response =~ /ten-leaf clover/) {
		    print "  You have a clover! Can\'t adventure in this area with this cloverprotect setting.\n";
		    $macroerror = 2 if ($macroerror < 2);
		    return 0;
		}
	    }
	    
	    # Equip the appropriate outfit. If the user has requested one
	    # of the disguised areas, then put on the appropriate disguise.
	    # If we're going to the peak and don't have a saucesphere,
	    # then we should equip the cold-weather gear.
	    # DO NOT equip anything if we're in the palindome -- we need
	    # to wear the Talisman o' Nam.
	    # Ideally this would be a little more configurable.
	    my $outfitsuccess = 1;
	    
	    if ($areanum == $areas{"icypeak"} && $hassphere==0) {
		# Perhaps we should recheck to see if the sphere is still there
		# from the last time we checked. If not, then it doesn't really
		# matter -- it'll get caught eventually.

		$outfitsuccess = &equip_outfit("coldweather");
	    } elsif ($areanum == $areas{"dfrathouse"}) {
		$outfitsuccess = &equip_outfit("fratboy");
	    } elsif ($areanum == $areas{"dhippycamp"}) {
		$outfitsuccess = &equip_outfit("hippy");
	    } elsif ($areanum == $areas{"dpiratecove"}) {
		$outfitsuccess = &equip_outfit("swashbuckling");
	    } elsif ($areanum == $areas{"dmine"}) {
		$outfitsuccess = &equip_outfit("mining");
	    } elsif ($areanum == $areas{"palindome"}) {
		print "Equipping Talisman...\n";
#		$outfitsuccess = do_action($verbs{"equip"}, $equipobjs{"talisman"}, 1);
	    } elsif ($alwaysequip == 1) {
		# Failing to equip the custom outfit shouldn't stop our macro.
		my $oldmacroerror = $macroerror;
		&equip_outfit($defaultoutfit);
		$macroerror = $oldmacroerror;
	    }
	    
	    # If we're random, we should really pick a new area rather than just go
	    # to that area without the outfit. But I'm not really picky.

	    if ($outfitsuccess == 0 && $areaname != "random" && $areaname != "arandom") {
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    }

	    print "Off to the $areanames{$areanum}...\n";
	}
    
	if ($onquest == 0) {
	    print "  Adventure $i / $nturns\n";
	} else {
	    print "  Adventure $i ($questcur/$questnum items so far)\n";
	}
	
	if ($areanum < 1) {
	    print "Unexpected fatal error in adventure(): areanum is $areanum\n";
	    $macroerror = 3 if ($macroerror < 3);
	    return 0;
	}

	&get_page("$site/adventure.php?adv=$areanum");
	
	# Check for fatal errors.
	
	if ($response =~ /You\'re out of adventures/) {
	    print "\tSorry, you're out of adventures.\n";
	    if ($nturns > 1) {
		print_advsummary();
	    }
	    $macroerror = 1 if ($macroerror < 1);
	    return 0;
	}
	if ($response =~ /way too beaten up/) {
	    print "\tYou're too beaten up to adventure.\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}
	if ($response =~ /extending your reach/) {
	    print "\tYou don't have any chewing gum on a string.\n";
	    $macroerror = 1 if ($macroerror < 1);
	    return 0;
	}
	if ($response =~ /faded back into the spectral mists/) {
	    print "\tThe Pickle Factory isn't around today.\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}
	if ($response =~ /shouldn\'t be here/) {
	    print "\tYou shouldn't be here.\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}
	if ($response =~ /(You must have at least .*? to adventure here\.)/) {
	    print "\t$1\n";
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}

	# Figure out if we're done or not.
	
	my $endbattle = 0;       # Flag to tell if we should keep fighting.
	my $firstpage = 1;       # Flag to indicate whether this is our first fight or not
	my $drunkenstupor = 0;   # Flag to tell if we fell into a drunken stupor.
	my $counter=0;           # Number of fight retries.

	if ($response =~ /<!--WINWINWIN-->/) {
	    &parse_battle_results($firstpage);
	    $endbattle = 1;
	}
	elsif ($response =~ /(Coyote Ugly|Gutterbound|Hey, baby\.  Wanna wrestle\?|The Too-Much Booze Blues|What\'s that smell\?)/) {
	    print "\tYou had a Drunken Stupor: $1\n";
	    $endbattle = 1;
	    $drunkenstupor = 1;
	}
	elsif ($response =~ /Adventure again/ || $response =~ /Go back to/) {
	    if ($response =~ /You slink away, dejected and defeated/) {
		# Ouch, a one-shot-kill of us.
		&parse_battle_results(1);
		print "  ** You lost! **\n";
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    }
	    elsif ($response =~ /<center><b>(.*?)<\/b><br>/) {
		print "\tYou had a noncombat adventure: $1\n";
	    }
	    else { print "\tYou didn't encounter an enemy.\n"; }
	    $endbattle = 1;
	}
	elsif ($response =~ /Denim Axes Examined/) {
	    # Special catch for Denim Axes Examined.
	    print "\tYou had a noncombat adventure: Denim Axes Examined\n";

	    # Politely decline the man's offer.
	    &get_page("-d \"pwd=$pwdhash&whichchoice=2&option=2\" $site/choice.php");
	    if ($response =~ /Okay, maybe some other time, then/ &&
		$response =~ /Adventure again/) {
		$endbattle = 1;
		print "\tYou politely decline the man's offer.\n";
		
		# Denim Axes Examined doesn't count as an adventure, so
		# set the counter back down.
		$i--;
	    }
	    else {
		print $response, "\n";
		print "Error: unexpected result in Denim Axes Examined, bailing...\n";
		$macroerror = 3 if ($macroerror < 3);
		return 0;
	    }
	}
	elsif ($response !~ /Combat/) {
	    # We're not in combat, but it's not because of one of the known cases, either.
	    # Bizarre. Print the result and bail.
	    print $response, "\n";
	    print "Alert: unexpected adventure result\n";
	    $endbattle = 1;
	}

	# We're in combat. Keep fighting until we stop.
	
	until ($endbattle==1) {
	    # Figure out why we didn't win. This only deals with fumbles
	    # at the moment but will deal with other stuff later.
	    
	    &parse_battle_results($firstpage);
	    $firstpage = 0;
	    
	    # If we're low on HP, run away! Otherwise, bravely fight.
	    if ($curhp < $fleewhen) {
		print "\t...HP down to $curhp, attempting to flee...\n";
		&get_page("-d \"action=runaway\" $site/fight.php");
	    }
	    else {
		print "\t...Continuing fight...\n";
		&get_page("-d \"action=$action\" $site/fight.php");
	    }
	    
	    $counter++;
	    if ($response =~ /You run away, like a sissy little coward/) {
		print "  You successfully ran away.\n";
		$macroerror = 1 if ($macroerror < 1);
		return 0;
	    }

	    if ($response =~ /You slink away, dejected and defeated/) {
		&parse_battle_results($firstpage);
		print "  ** You lost! **\n";
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    }
	    
	    # If we try to run away but fail, then we don't get the
	    # "you slink away, dejected and defeated" message. So we
	    # need a different check.
	    
	    if ($response =~ /You attempt to run away like a sissy little coward, but you don\'t quite make it\./ &&
		($response =~ /Adventure again/ || $response =~ /Go back to/)) {
		&parse_battle_results($firstpage);
		print "  ** You lost! **\n";
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    }

	    # Check to make sure that we haven't fallen back to the main
	    # map. This shouldn't actually happen but if it does we
	    # should stop rather than keep trying.

	    if ($response =~ /town\.php/ && $response =~ /campground\.php/
		&& $response =~ /plains\.php/) {
		print "   Unexpected end of fight! Ending adventure...\n";
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    }

	    if ($counter > 30) {
		&parse_battle_results($firstpage);
		print "Max fight retries exceeded! Giving up.\n";
		$macroerror = 2 if ($macroerror < 2);
		return 0;
	    }
	    if ($response =~ /<!--WINWINWIN-->/) {
		&parse_battle_results($firstpage);
		$endbattle = 1;
	    }
	}
	
	&parse_response();
	
	# If we had a drunken stupor and didn't want to,
	# get out of here!
	if ($drunkenstupor == 1 && $areaname ne "stupor") {
	    $macroerror = 2 if ($macroerror < 2);
	    return 0;
	}

	# If we've finished our quest, yay!
	if ($onquest == 1 && $questnum <= $questcur) {
	    print "Quest finished!\n";
	    last;
	}

#  Healevery functionality. This is commented out for the time being
#  until I actually finish implementing it.
#
#	if ($healevery > 0 && ($i % $healevery) == 0) {
#	    print "Healing yourself...\n";
#	    &get_page("$healcommand $site/$healpage");
#	}
    }
    
    if ($onquest == 1 && $questcur < $questnum) { print "Reached quest maximum of $questmax but only got $questcur of requested item.\n"; }
    if ($areaname eq "sewer") { print "You got $n worthless item", ($n == 1) ? "" : "s", ".\n"; }
    
    # Print out the total items gained.
    
    if ($nturns > 1) {
	print_advsummary();
    }
    return 1;
}

sub get_allowedobjects($) {
    $allowedobjects = 0;
    my $verb = shift;
    my $promptword = "<item>";

    # This could possibly be done with symbolic references rather than setting it by
    # hand, but they're scary. So by hand it is.

    if ($verb eq "adventure") { $allowedobjects = \%areas; $promptword = "<destination>"; }
    elsif ($verb eq "buff") { $allowedobjects = \%buffobjs; $promptword = "<buff> <target>"; }
    elsif ($verb eq "buy") { $allowedobjects = \%buyobjs; }
    elsif ($verb eq "clan") { $allowedobjects = \%clanobjs; $promptword = "<action>"; }
    elsif ($verb eq "campground") { $allowedobjects = \%campgroundobjs; $promptword = "<action>"; }
    elsif ($verb eq "donate") { $allowedobjects = \%donateobjs; $promptword = "<shrine>"; }
    elsif ($verb eq "drink") { $allowedobjects = \%drinkobjs; }
    elsif ($verb eq "eat") { $allowedobjects = \%eatobjs; }
    elsif ($verb eq "equip") { $allowedobjects = \%equipobjs; }
    elsif ($verb eq "execute") { $allowedobjects = \%macros; $promptword = "<macro>"; }
    elsif ($verb eq "familiar") { $allowedobjects = \%familiarobjs; $promptword = "<familiar>"; }
    elsif ($verb eq "getbuff") { $allowedobjects = \%getbuffobjs; $promptword = "<buff>"; }
    elsif ($verb eq "gym") { $allowedobjects = \%gymobjs; $promptword = "<attribute>"; }
    elsif ($verb eq "halloween") { $allowedobjects = \%halloweenobjs; }
    elsif ($verb eq "help") { $allowedobjects = \%verbs; $promptword = "<command>"; }
    elsif ($verb eq "hermit") { $allowedobjects = \%hermititems; }
    elsif ($verb eq "junk") { $allowedobjects = \%junkobjs; $promptword = "<area>"; }
    elsif ($verb eq "mallbuy") { $allowedobjects = \%mallobjs; }
    elsif ($verb eq "mix") { $allowedobjects = \%mixobjs; $promptword = "<drink>"; }
    elsif ($verb eq "outfit") { $allowedobjects = \%outfitobjs; $promptword = "<outfit>"; }
    elsif ($verb eq "send") { $allowedobjects = \%sendobjs; $promptword = "<item> <target>"; }
    elsif ($verb eq "set") { $allowedobjects = \%setobjs; $promptword = "<parameter>"; }
    elsif ($verb eq "shore") { $allowedobjects = \%shoreobjs; $promptword = "<destination>"; }
    elsif ($verb eq "skill") { $allowedobjects = \%skillobjs; $promptword = "<skill>"; }
    elsif ($verb eq "use") { $allowedobjects = \%useobjs; }
    # Nothing gets set for moons, status, or trinket because they don't take objects.

    return $promptword;
}

sub do_help($) {
    my $verb = shift;

    if ($verb eq "all") {
	print "Welcome to KoL CLI $version. Type help <command> for help on a specific command.\n";
	print "All commands may be abbreviated as long as the abbreviation is not ambiguous.\n";
	print "All commands default to 1 turn or item if no number is specified.\n";
	print "Available commands are:\n";
	foreach my $verb (sort keys %verbhelp) { print " ", $verbhelp{$verb}, "\n"; }
    }
    else {
	print $verbhelp{$verb}, "\n";
	&get_allowedobjects($verb);

	if ($allowedobjects) {
	    print "The following arguments are supported:\n";
	    if ($verb eq "adventure") {
		foreach my $item (sort keys %$allowedobjects) {
		    print $item, " " x (19-length($item)), $areanames{$areas{$item}}, "\n";
		}
	    }
	    elsif ($verb eq "set") {
		foreach my $item (sort keys %$allowedobjects) {
		    print $item, " " x (15-length($item)), $sethelp{$item}, "\n";
		}
	    }
	    else {
		prettyprint("", sort keys %$allowedobjects);
	    }
	}
	elsif ($verb eq "trinket") {
	    print "This command will repeatedly buy gum and adventure in the sewer until <n>\n";
	    print "trinkets have been obtained.\n";
	}

	if ($verb eq "outfit" && %customoutfits) {
	    print "Your character also has the following known Custom Outfits:\n";
	    prettyprint("", sort keys %customoutfits);
	}
    }
}

# Read the configuration file in.

sub read_configfile() {
    my $configfile = glob("~/.kolconfig");

    # If the expansion fails, then just go to the current directory.

    if (!$configfile || $configfile =~ /^~/) {
	$configfile = ".kolconfig";
    }
    if (open CONFIGFILE, $configfile) {
	my $curuser;
	my $curmacro;
	my $curstate = 0;               # 1 if we're inside a macro

	while (my $line = <CONFIGFILE>) {
	    # Preprocessing.
	    chomp $line;
	    $line =~ s/^ *//;
	    $line =~ s/ *$//;
	    $line =~ s/#.*// unless $line =~ /^pass/i;
	    next if ($line =~ /^ *$/);  # skip lines that are still blank

	    my @linewords = split(" ", $line);
	    
	    next unless @linewords;     # shouldn't happen but in case it does

	    if ($linewords[0] =~ /^user$/i) {
		if (@linewords) { $curuser = join(" ", @linewords[1..$#linewords]);
				  push @users, $curuser;}
		else { print "Error in config file: user specified with no argument\n"; }
	    }
	    elsif ($linewords[0] =~ /^pass$/i) {
		if (@linewords) { $passwords{$curuser} = join(" ", @linewords[1..$#linewords]); }
		else { print "Error in config file: pass specified with no argument\n"; }
	    }
	    elsif ($linewords[0] =~ /^pwdhash$/i) {
		print "Warning in config file: pwdhash is no longer supported\n";
	    }
	    elsif ($linewords[0] =~ /^macro$/i) {
		if ($linewords[1]) {
		    $curmacro = $linewords[1];
		    $curstate = 1;
		}
		else { print "Error in config file: macro specified with no argument\n"; }
	    }
	    elsif ($linewords[0] =~ /^set$/i && $curstate == 0) {
		# Settings. First make sure we have a valid parameter.

		if ($linewords[1] && $setobjs{$linewords[1]}) {
		    my $param = $linewords[1];
		    my $value = 1;
		    # We use exists here because otherwise it'll fail to register 0.
		    if (exists($linewords[2])) {
			# If there is no $setvalues, just make sure it's numeric.
			# Otherwise, make sure it's in the allowed set.
			if ((!$setvalues{$param} && $linewords[2] =~ /^(\d+)$/) ||
			    ($setvalues{$param} && grep /^$linewords[2]$/, @ {$setvalues{$param}}) ) {
			    $value = $linewords[2];
			}
			else {
			    print "Error in config file: bad value $linewords[2] for set $param\n";
			    next;
			}
		    }
		    
		    # Now check to see if we have a user or not. If not, make the setting
		    # immediately. If so, then stick it in the settings hash.
		    if (!$curuser) {
			$ {$setobjs{$param}} = $value;
		    }
		    else {
			$settings{$curuser}{$param} = $value;
		    }
		}
		else {
		    print "Error in config file: set specified with no argument\n" if (!$linewords[1]);
		    print "Error in config file: set specified with bad argument $linewords[1]\n" if ($linewords[1]);
		}
	    } # set
	    elsif ($curstate == 1) {
		push @ {$macros{$curmacro}}, $line;
	    }
	    else { print "Error: unrecognized command in config file $line\n"; }
	} # line loop
    } # if there was a config file

}

# End of subroutines -- main program begins here

# Read in config file.

&read_configfile();

# Read in extensions.

if (open EXTFILE, "extensions.kol") {
    my @extensions = <EXTFILE>;
    unless (eval (join (" ", @extensions))) {
	die "Error in reading in extensions file: $@\n";
    }
}

# Look at command line arguments. Anything is interpreted
# as a username except -b, which is the "batch mode" flag
# and will create a nice macro which will execute the
# rest of the command line.

while (my $argument = shift) {
    if ($argument =~ /-(.)/) {
	if ($1 eq "b") {
	    print "Entering batch mode...\n";
	    my $argument = join(" ", @ARGV);
	    @currentmacro = ($argument, "quit");
	    $batchmode = 1;
	    $macroexecuting = 1;
	    last;
	}
	else { die "Unknown flag -$1!\n"; }
    }
    else { $username = $argument; }
}

# If we have a config file, and there's only one username
# in it, assume that we want that username.

if (!$username && $#users == 0) {
    $username = $users[0];
}

print "KoL CLI v $version\n";
print " (c) 2004\n";
print " Kingdom of Loathing (c) 2004 Asymmetric Publications, LLC\n";

# If no username, get it from the terminal.

if (!$username) {
    print "Login: ";
    $username = <STDIN>;
    chomp $username;
}

print "Began executing at ", scalar localtime time(), "\n" if $batchmode;

&start_new_user();

# Main loop.

MAINLOOP: while (1) {
    # If we're executing a macro, check to see if we should stop or abort.

    if ($macroexecuting == 1) {
	# If we're done with the macro, stop executing.
	if (!@currentmacro) { $macroexecuting = 0; }
    
	# Check to see if the current macro error level is greater than
	# our tolerance for them.
	# 3 = syntax errors
	# 2 = generally fatal errors
	# 1 = generally nonfatal errors
	# 0 = no error
	
	if ($macroerror > $macrofail) {
	    print "Error in macro, stopping macro execution...\n";
	    if ($batchmode == 0) {
		@currentmacro = ();
		$macroexecuting = 0;
	    }
	    else {
		# If we're in batch mode, we want to quit now.
		# Otherwise we'll just hang.
		@currentmacro = ("quit");
	    }
	}
    }

    # Reset $macroerror and $onquest.
    $macroerror = 0;
    $onquest = 0;
	
    my $input;

    print "> ";

    # If we're running a macro, take the next line from it. Otherwise, get
    # it from stdin.

    if ($macroexecuting == 0) {
	$input = <STDIN>;
    } else {
	$input = shift @currentmacro;
	print $input, "\n";
    }

    # Check to see if we got an EOF. If so, logout. Note that this step comes BEFORE
    # the chomp so that we can distinguish lines which are just empty from EOFs.

    unless ($input) {
	$input = "logout";
	print $input, "\n";
    }

    chomp $input;

    # Parse the input line. First make sure there's something in it.

    if ($input =~ /^ *$/) {
	print "Pardon?\n";
	next;
    }
    
    # Strip comments.

    $input =~ s/#.*//;
    next if ($input =~ /^ *$/);  # Continue silently if this line was all comment.
    
    # If we fail AFTER here, then we have a syntax error in our input line.
    # Set macroerror appropriately.
    $macroerror = 3;

    # Now check to see if the verb is something we know.

    my @inputwords = split(" ", $input);
    next unless @inputwords;    # This shouldn't happen but in case it does, continue quietly.
    my @potentialverbs = grep (/^\Q$inputwords[0]/i, keys %verbs);
    if (@potentialverbs == 0) {
	print "Unknown command: $inputwords[0]\n";
	next;
    }
    if (@potentialverbs > 1) {
	print "Ambiguous command: ", join(", ", @potentialverbs), "\n";
	next;
    }

    my $verb = $potentialverbs[0];
    shift @inputwords;

    # If the user wants to quit, do that.

    if ($verb eq "quit" || $verb eq "logout" || $verb eq "exit" || $verb eq "switch") {
	# Note that the logout does NOT go through get_page. This is because we
	# don't really care if it succeeds or fails, and we certainly don't want
	# to relogin just to log out (not that I think this happens anyway).

	print "Logging out... ";
	@response = `curl $options $site/logout.php` unless $noconnect;
	$response = join(" ", @response);
	if ($response =~ /You have been logged out/) {
	    print "You are now logged out.\n";
	}

	print "Finished executing at ", scalar localtime time(), "\n" if $batchmode;

	# Kill the cookie file.
	unlink(".kolcookies-$cookiefilename");

	if ($verb ne "switch") {
	    print "Thank you for using KoL CLI $version.\n";
	    exit;
	}
    }

    # Repeat functionality is added in here. This is highly sneaky
    # and more than a little hackish.

    if ($verb eq "repeat") {
	if (!$prevverb) {
	    print "Nothing to repeat!\n";
	    next MAINLOOP;
	}
	$verb = $prevverb;
	@inputwords = @prevwords;
    }
    $prevverb = $verb;
    @prevwords = @inputwords;

    # Now, parse the rest of the line. All lines are of the form <verb> <object>
    # [<optional number>] (with a couple of exceptions).
    
    my $number = 1;

    # get_allowedobjects sets $allowedobjects and returns $promptword.
    
    my $promptword = &get_allowedobjects($verb);

    if ($allowedobjects) {
	my $argument = shift @inputwords;

	# If the argument has quotes, keep adding until we run out of quotes.
	# Really this should be done in the split up there rather than
	# individually. However there's enough special cases to make that
	# really annoying (namely, we want to strip quotes here, but not
	# from things like quest items, because otherwise we have no way
	# of distinguishing "334" from 334).

	while ($argument =~ /^\"/ && $argument !~ /\"$/) {
	    if (!@inputwords) {
		print "Error: no close quote found!\n";
		next MAINLOOP;
	    }
	    $argument .= " " . shift @inputwords;
	}
	$argument =~ s/\"//g;

	unless ($argument) {
	    if ($verb eq "help") { &do_help("all"); }
	    elsif ($verb eq "set") { foreach my $item (sort keys %setobjs) { print "$item: $ {$setobjs{$item}}\n"; } }
	    else { print "Usage: $verb $promptword [<number>]\n"; }
	    next;
	}
	my @potentialobjects = grep (/^\Q$argument/i, keys %$allowedobjects);

	if (@potentialobjects == 0) {
	    if ($verb eq "outfit") {
		# In this case, it's okay to have an unknown object
		# -- equip_outfit() will take care of it.
		$potentialobjects[0] = $argument;
	    }
	    else {
		print "Unknown argument for $verb: $argument\n";
		next;
	    }
	}
	if (@potentialobjects > 1) {
	    # Check to see if we have an EXACT match. In this case, use it.
	    my @exactmatches = grep (/^\Q$argument\E$/i, keys %$allowedobjects);
	    if (@exactmatches == 1) {
		@potentialobjects = @exactmatches;
	    }
	    else {
		print "Ambiguous arguments for $verb: ", join(", ", @potentialobjects), "\n";
		next;
	    }
	}
	$object = $potentialobjects[0];
    } # if this requires an object

    # For "buff", "switch", and "send", we need to find a target too. This one is tricky.

    my $target = $username;
    my $target_is_number = 0;
    my $message;

    if ($verb eq "buff" || $verb eq "switch" || $verb eq "send") {
	if (!@inputwords) {
	    print "Usage: $verb $promptword [<number>]\n";
	    next;
	}
	$target = shift @inputwords;
	
	# If the target is entirely numeric, assume that this is a player ID, not a player name.

	if ($target =~ /^\d+$/) { $target_is_number = 1; }
	
	# If the target contains quotes, we should keep adding words until we find a close quote.

	while ($target =~ /^\"/ && $target !~ /\"$/) {
	    if (!@inputwords) {
		print "Error: no close quote found!\n";
		next MAINLOOP;
	    }
	    $target .= " " . shift @inputwords;
	}

	# Clean the target.
	$target =~ tr/A-Za-z0-9_\- //cd;
    }

    # Check to see if the user gave us a number (or other special value
    # that could go in this argument).
    
    if (@inputwords) {
	my $argument = shift @inputwords;

	if ($argument eq "n") { $number = $n; }
	elsif ("until" =~ /^\Q$argument/i) {
	    if (!@inputwords) {
		print "Error: Expected an argument after \"until\" but didn't get one!\n";
		next MAINLOOP;
	    }

	    my $untilargument = shift @inputwords;
	    
	    # There's two possible possibilities here: quest mode or regular
	    # mode.

	    if ($untilargument =~ /^(\d+)/) {
		$untilargument = $1;

		# This will abort the instruction (leaving $macroerror==4) if 
		# &get_charpane() fails for some reason. This is probably the
		# safest tack to take. Otherwise, we might end up adventuring
		# way too much.
		
		&get_charpane(0) or next;
		$number = ($advleft > $untilargument) ? ($advleft - $untilargument) : 0;
		if ($verb eq "shore") { $number = int($number/3); }
		print "Will adventure for $number turns.\n";
	    }
	    elsif ($verb eq "adventure") {
		# Can only quest on adventures for now.
		
		$onquest=1;
		$questcur=0;
		$questobj = $untilargument;

		# If we found an open quote, keep going until
		# we get a close quote.
		
		while ($questobj =~ /^\"/ && $questobj !~ /\"$/) {
		    if (!@inputwords) {
			print "Error: no close quote found!\n";
			next MAINLOOP;
		    }
		    $questobj .= " " . shift @inputwords;
		}
		$questobj =~ s/\"//g;

		if (@questobjs && !(grep /\Q$questobj/, @questobjs)) {
		    print "Error: $questobj is not in the list of acceptable quest objects.\n";
		    prettyprint("Acceptable quest objects", @questobjs);
		    next MAINLOOP;
		}
		
		# See if we got a number of objects. Otherwise, use 1.

		if (@inputwords) {
		    my $questargument = shift @inputwords;
		    if ($questargument =~ /^(\d+)/) {
			$questnum = $1;
		    }
		    else {
			print "Error: Expected a number of objects to quest for!\n";
			next MAINLOOP;
		    }
		}
		else { $questnum = 1; }

		$number = $questmax;
		$number = 250 if ($number == 0);
	    } # quest mode
	    else {
		print "Error: Expected a number after \"until\" but didn't get one!\n";
		next MAINLOOP;
	    } # error
	}
	elsif ($argument =~ /^(\d+)%/) {
	    $number = $1;
	    
	    # See comment above.
	    &get_charpane(0) or next;
	    if ($verb eq "skill" || $verb eq "buff") {
		if (!$mpcosts{$object}) {
		    print "Error! Don't know how many MP $object costs!\n";
		    next;
		}
		$number = int(($number*$curmp)/(100*$mpcosts{$object}));
		print "Will cast $number times (using ", $mpcosts{$object}*$number, " MP).\n";
	    }
	    elsif ($verb eq "adventure" || $verb eq "shore") {
		$number = int($advleft*$number/100);
		$number = int($number/3) if $verb eq "shore";
		print "Will adventure for $number turns.\n";
	    }
	    else {
		print "Can only use % arguments for adventure, shore, skill, or buff.\n";
		next;
	    }
	}
	elsif ($argument =~ /^(\d+)/) { $number = $1; }
	else { 
	    print "Error: Expected a number but got $argument\n";
	    next;
	}
	
	# If there's anything else, throw it away, unless it's being put in
	# a message.

	if (@inputwords) {
	    if ($verb eq "send") {
		$message = join(" ", @inputwords);
		# Properly encode the messsage.
		$message =~ s/([^-A-Za-z0-9_])/sprintf("%%%02X", ord($1))/ge;
	    }
	    else { print "Input discarded: @inputwords\n"; }
	}
    }

    my $targetpage = $verbs{$verb};

    # If we have an object that can be used multiply, then we want to call
    # multiuse.php, NOT inv_use.php.

    if ($verb eq "use" && $useobjs{$object} =~ /useitem/) {
	$targetpage = "multiuse.php";
    }

    # If we got this far, then we DON'T have a syntax error in our macro
    # (with one exception, in set below). Let the called routine set
    # it.

    $macroerror = 0;

    # If the number specified is 0, then we shouldn't do anything -- unless
    # we're sending something, in which case we might want to send 0, because
    # we might want a message. In any case, this isn't an error.

    if ($number == 0 && $verb ne "send") {
	print "No actions taken.\n";
	next MAINLOOP;
    }

    # Check to see if the buffbots are online.

    if ($verb eq "getbuff" && $checkbuffbots == 1) {
	# If we don't know their status, then find out.
	# I only check Petal but I assume the two are the same.
	if (!defined($buffbots)) {
	    print "Checking buffbot status... ";
	    &get_page("-d \"who=98842\" $site/showplayer.php");
	    if ($response =~ /Offline/) {
		print "Buffbots are offline.\n";
		$buffbots = 0;
	    }

	    # I've tried to make this as general as possible
	    # given that I don't remember what the actual
	    # syntax is. If/when the buffbots ever come back
	    # then I'll make this check better.
	    elsif ($response =~ /Empathy/i) {
		print "Buffbots online!\n";
		$buffbots = 1;
	    }
	    else {
		print "Couldn't tell whether the buffbots are online!\n";
		$buffbots = 0;
	    }
	}
	if ($buffbots == 0){
	    print "Buffbots are currently offline. Meat will not be sent.\n";
	    print "If you want to send meat anyway, set checkbuffbots to 0.\n";
	    next MAINLOOP;
	}
    }

    # Finally, call the appropriate subroutine.

    if ($verb eq "adventure") { &adventure($object, $number); }
    elsif ($verb eq "hermit") { &visit_hermit($object, $number); }
    elsif ($verb eq "trinket") { &get_trinket($number); }
    elsif ($verb eq "junk" ) { &sell_junk($object); }
    elsif ($verb eq "outfit") { &equip_outfit($object); }
    elsif ($verb eq "set") {
	if (!$setvalues{$object} || grep /$number/, @ {$setvalues{$object}}) {
	    $ {$$allowedobjects{$object}} = $number;
	    print "Set $object to $number";
	    if ($object ne "cloverprotect" && $object ne "macrofail" && $object ne "questmax" &&
		$object ne "maxdrunk" && $object ne "fleewhen") {
		print " (", $number ? "on" : "off", ")";
	    }
	    if (($object eq "questmax" || $object eq "fleewhen") && $number == 0) { print " (off)"; }
	    print ".\n";
	}
	else {
	    $macroerror = 4;
	    print "Bad value for $object: $number (allowed: ", join(", ", @ {$setvalues{$object}}), ")\n";
	}
    }
    elsif ($verb eq "help") { &do_help($object); }
    elsif ($verb eq "execute") {
	my $subs = 0;
	$macroexecuting = 1;
	my @addmacro = @ {$macros{$object}};
	$subs += s/ arg$/ $number/i foreach @addmacro;
	
	# If the macro contains "arg", add it once.
	# Otherwise, add it that number of times.
	
	if ($subs != 0) { $number = 1; }
	@currentmacro = ((@addmacro) x $number, @currentmacro);
    }
    elsif ($verb eq "familiar") { &equip_familiar($object); }
    elsif ($verb eq "status") { &get_charpane(1); }
    elsif ($verb eq "moons") { &get_moonpane(1); }
    elsif ($verb eq "togglemode") { &mode_toggle(); }
    elsif ($verb eq "switch") { 
	$username = $target;
	&start_new_user();
    }
    else {
	my $params = $$allowedobjects{$object};
	# Put in the name obtained from parsing.
	if ($params =~ /NAME/) {
	    # Buffs require special handling because there are two fields for
	    # the player to go in.
	    if ($verb eq "buff") {
		if ($target_is_number == 1) {
		    $params =~ s/NAME/&targetplayer=$target/;
		}
		else {
		    # Do a quick encode of the URL.
		    $target =~ s/ /%20/g;
		    $params =~ s/NAME/$target&targetplayer=0/;
		}
	    }
	    else {
		$target =~ s/ /%20/g;
		$params =~ s/NAME/$target/;
	    }
	} # name substitution in
	
	# If we're sending a message, add it in.
	if ($verb eq "send" && $message) { $params .= $message; }

	&do_action($targetpage, $params, $number);
    }
}
